# Copyright (c) 2017 Fujitsu Limited
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

from neutron_lib import constants as lib_const
from neutron_lib.services.logapi import constants as log_const
from os_ken.base import app_manager
from os_ken.lib.packet import packet
from oslo_config import cfg
from oslo_log import formatters
from oslo_log import handlers
from oslo_log import log as logging

from neutron.agent.linux.openvswitch_firewall import constants as ovsfw_consts
from neutron.agent.linux.openvswitch_firewall import firewall as ovsfw
from neutron.agent.linux.openvswitch_firewall import rules
from neutron.plugins.ml2.drivers.openvswitch.agent.common import constants \
        as ovs_consts
from neutron.services.logapi.agent import log_extension as log_ext
from neutron.services.logapi.common import exceptions as log_exc
from neutron.services.logapi.drivers.openvswitch import log_oskenapp

LOG = logging.getLogger(__name__)

OVS_FW_TO_LOG_TABLES = {
    ovs_consts.RULES_EGRESS_TABLE: ovs_consts.ACCEPTED_EGRESS_TRAFFIC_TABLE,
    ovs_consts.RULES_INGRESS_TABLE: ovs_consts.ACCEPTED_INGRESS_TRAFFIC_TABLE,
}

FIELDS_TO_REMOVE = ['priority', 'actions', 'dl_type',
                    'reg_port', 'reg_remote_group']

REMOTE_RULE_PRIORITY = 70


def setup_logging():
    log_file = cfg.CONF.network_log.local_output_log_base
    if log_file:
        # pylint: disable=import-outside-toplevel
        from logging import handlers as watch_handler
        log_file_handler = watch_handler.WatchedFileHandler(log_file)
        log_file_handler.setLevel(
            logging.DEBUG if cfg.CONF.debug else logging.INFO)
        LOG.logger.addHandler(log_file_handler)
        log_file_handler.setFormatter(
            formatters.ContextFormatter(
                fmt=cfg.CONF.logging_default_format_string,
                datefmt=cfg.CONF.log_date_format))
    elif cfg.CONF.use_journal:
        journal_handler = handlers.OSJournalHandler()
        LOG.logger.addHandler(journal_handler)
    else:
        syslog_handler = handlers.OSSysLogHandler()
        LOG.logger.addHandler(syslog_handler)


def find_deleted_sg_rules(old_port, new_ports):
    del_rules = list()
    for port in new_ports:
        if old_port.id == port.id:
            for rule in old_port.secgroup_rules:
                if rule not in port.secgroup_rules:
                    del_rules.append(rule)
            return del_rules
    return del_rules


class Cookie(object):

    def __init__(self, cookie_id, port, action, project):
        self.id = cookie_id
        self.port = port
        self.action = action
        self.project = project
        self.log_object_refs = set()

    def __eq__(self, other):
        return (self.id == other.id and
                self.action == other.action and
                self.port == other.port)

    def __hash__(self):
        return hash(self.id)

    def add_log_obj_ref(self, log_id):
        self.log_object_refs.add(log_id)

    def remove_log_obj_ref(self, log_id):
        self.log_object_refs.discard(log_id)

    @property
    def is_empty(self):
        return not self.log_object_refs


class OFPortLog(object):

    def __init__(self, port, ovs_port, log_event):
        self.id = port['port_id']
        self.ofport = ovs_port.ofport
        self.secgroup_rules = [self._update_rule(rule) for rule in
                               port['security_group_rules']]
        # event can be ALL, DROP and ACCEPT
        self.event = log_event

    def _update_rule(self, rule):
        protocol = rule.get('protocol')
        if protocol is not None:
            if not isinstance(protocol, int) and protocol.isdigit():
                rule['protocol'] = int(protocol)
            elif (rule.get('ethertype') == lib_const.IPv6 and
                  protocol == lib_const.PROTO_NAME_ICMP):
                rule['protocol'] = lib_const.PROTO_NUM_IPV6_ICMP
            else:
                rule['protocol'] = lib_const.IP_PROTOCOL_MAP.get(
                    protocol, protocol)
        return rule


class OVSFirewallLoggingDriver(log_ext.LoggingDriver):

    SUPPORTED_LOGGING_TYPES = ['security_group']
    REQUIRED_PROTOCOLS = [
        ovs_consts.OPENFLOW13,
        ovs_consts.OPENFLOW14,
    ]

    def __init__(self, agent_api):
        integration_bridge = agent_api.request_int_br()
        self.int_br = self.initialize_bridge(integration_bridge)
        self._deferred = False
        self.log_ports = collections.defaultdict(dict)
        self.cookies_table = set()
        self.cookie_ids_to_delete = set()
        self.conj_id_map = ovsfw.ConjIdMap(self.int_br.br)

    def initialize(self, resource_rpc, **kwargs):
        self.resource_rpc = resource_rpc
        setup_logging()
        self.start_logapp()

    @staticmethod
    def initialize_bridge(bridge):
        bridge.add_protocols(*OVSFirewallLoggingDriver.REQUIRED_PROTOCOLS)
        # set rate limit and burst limit for controller
        bridge.set_controller_rate_limit(cfg.CONF.network_log.rate_limit)
        bridge.set_controller_burst_limit(cfg.CONF.network_log.burst_limit)
        return bridge.deferred(full_ordered=True)

    def start_logapp(self):
        app_mgr = app_manager.AppManager.get_instance()
        self.log_app = app_mgr.instantiate(log_oskenapp.OVSLogOSKenApp)
        self.log_app.start()
        self.log_app.register_packet_in_handler(self.packet_in_handler)

    def packet_in_handler(self, ev):
        msg = ev.msg
        cookie_id = msg.cookie
        pkt = packet.Packet(msg.data)
        try:
            cookie_entry = self._get_cookie_by_id(cookie_id)
            LOG.info("action=%s project_id=%s log_resource_ids=%s vm_port=%s "
                     "pkt=%s", cookie_entry.action, cookie_entry.project,
                     list(cookie_entry.log_object_refs),
                     cookie_entry.port, pkt)
        except log_exc.CookieNotFound:
            LOG.warning("Unknown cookie=%s packet_in pkt=%s", cookie_id, pkt)

    def defer_apply_on(self):
        self._deferred = True

    def defer_apply_off(self):
        if self._deferred:
            self.int_br.apply_flows()
            self._cleanup_cookies()
            self._deferred = False

    def _get_cookie(self, port_id, action):
        for cookie in self.cookies_table:
            if cookie.port == port_id and cookie.action == action:
                return cookie

    def _get_cookies_by_port(self, port_id):
        cookies_list = []
        for cookie in self.cookies_table:
            if cookie.port == port_id:
                cookies_list.append(cookie)
        return cookies_list

    def _get_cookie_by_id(self, cookie_id):
        for cookie in self.cookies_table:
            if str(cookie.id) == str(cookie_id):
                return cookie
        raise log_exc.CookieNotFound(cookie_id=cookie_id)

    def _cleanup_cookies(self):
        cookie_ids = self.cookie_ids_to_delete
        self.cookie_ids_to_delete = set()
        for cookie_id in cookie_ids:
            self.int_br.br.unset_cookie(cookie_id)

    def generate_cookie(self, port_id, action, log_id, project_id):
        cookie = self._get_cookie(port_id, action)
        if not cookie:
            cookie_id = self.int_br.br.request_cookie()
            cookie = Cookie(cookie_id=cookie_id, port=port_id,
                            action=action, project=project_id)
            self.cookies_table.add(cookie)
        cookie.add_log_obj_ref(log_id)
        return cookie.id

    def _schedule_cookie_deletion(self, cookie):
        # discard a cookie object
        self.cookies_table.remove(cookie)
        # schedule to cleanup cookie_ids later
        self.cookie_ids_to_delete.add(cookie.id)

    def start_logging(self, context, **kwargs):
        LOG.debug("start logging: %s", str(kwargs))
        for resource_type in self.SUPPORTED_LOGGING_TYPES:
            # handle port updated, agent restarted
            if 'port_id' in kwargs:
                self._handle_logging('_create', context,
                                     resource_type, **kwargs)
            else:
                self._handle_log_resources_by_type(
                    '_create', context, resource_type, **kwargs)

    def stop_logging(self, context, **kwargs):
        LOG.debug("stop logging: %s", str(kwargs))
        for resource_type in self.SUPPORTED_LOGGING_TYPES:
            # handle port deleted
            if 'port_id' in kwargs:
                self._handle_logging('_delete', context,
                                     resource_type, **kwargs)
            else:
                self._handle_log_resources_by_type(
                    '_delete', context, resource_type, **kwargs)

    def _handle_log_resources_by_type(
            self, action, context, resource_type, **kwargs):

        log_resources = []
        for log_obj in kwargs.get('log_resources', []):
            if log_obj['resource_type'] == resource_type:
                log_resources.append(log_obj)
        if log_resources:
            self._handle_logging(
                action, context, resource_type, log_resources=log_resources)

    def _handle_logging(self, action, context, resource_type, **kwargs):
        handler_name = "%s_%s_log" % (action, resource_type)
        handler = getattr(self, handler_name)
        handler(context, **kwargs)

    def create_ofport_log(self, port, log_id, log_event):
        port_id = port['port_id']
        ovs_port = self.int_br.br.get_vif_port_by_id(port_id)
        if ovs_port:
            of_port_log = OFPortLog(port, ovs_port, log_event)
            self.log_ports[log_id].add(of_port_log)

    def _create_security_group_log(self, context, **kwargs):

        port_id = kwargs.get('port_id')
        log_resources = kwargs.get('log_resources')
        logs_info = []
        if port_id:
            # try to clean port flows log for port updated/create event
            self._cleanup_port_flows_log(port_id)
            logs_info = self.resource_rpc.get_sg_log_info_for_port(
                context,
                resource_type=log_const.SECURITY_GROUP,
                port_id=port_id)
        elif log_resources:
            logs_info = self.resource_rpc.get_sg_log_info_for_log_resources(
                context,
                resource_type=log_const.SECURITY_GROUP,
                log_resources=log_resources)

        for log_info in logs_info:
            log_id = log_info['id']
            old_ofport_logs = self.log_ports.get(log_id, [])
            ports = log_info.get('ports_log')
            self.log_ports[log_id] = set()
            for port in ports:
                self.create_ofport_log(port, log_id, log_info.get('event'))

            # try to clean flows log if sg_rules are deleted
            for port in old_ofport_logs:
                del_rules = find_deleted_sg_rules(
                    port, self.log_ports[log_id])
                if del_rules:
                    self._delete_sg_rules_flow_log(port, del_rules)

            for port_log in self.log_ports[log_id]:
                self.add_flows_from_rules(port_log, log_info)

    def _cleanup_port_flows_log(self, port_id):
        cookies_list = self._get_cookies_by_port(port_id)
        for cookie in cookies_list:
            if cookie.action == log_const.ACCEPT_EVENT:
                self._delete_flows(
                    table=ovs_consts.ACCEPTED_INGRESS_TRAFFIC_TABLE,
                    cookie=cookie.id)
                self._delete_flows(
                    table=ovs_consts.ACCEPTED_EGRESS_TRAFFIC_TABLE,
                    cookie=cookie.id)
            if cookie.action == log_const.DROP_EVENT:
                self._delete_flows(
                    table=ovs_consts.DROPPED_TRAFFIC_TABLE,
                    cookie=cookie.id)
            self._schedule_cookie_deletion(cookie)

    def _delete_security_group_log(self, context, **kwargs):
        # port deleted event
        port_id = kwargs.get('port_id')

        if port_id:
            self._cleanup_port_flows_log(port_id)

        # log resources deleted events
        for log_resource in kwargs.get('log_resources', []):
            log_id = log_resource.get('id')
            of_port_logs = self.log_ports.get(log_id, [])
            for of_port_log in of_port_logs:
                self.delete_port_flows_log(of_port_log, log_id)

    def _log_accept_flow(self, **flow):
        # log first accepted packet
        flow['table'] = OVS_FW_TO_LOG_TABLES[flow['table']]
        flow['actions'] = 'controller'
        # forward egress accepted packet and log
        if flow['table'] == ovs_consts.ACCEPTED_EGRESS_TRAFFIC_TABLE:
            flow['actions'] = 'resubmit(,%d),controller' % (
                ovs_consts.ACCEPTED_EGRESS_TRAFFIC_NORMAL_TABLE)
        self._add_flow(**flow)

    def _add_flow(self, **kwargs):
        dl_type = kwargs.get('dl_type')
        ovsfw.create_reg_numbers(kwargs)
        if isinstance(dl_type, int):
            kwargs['dl_type'] = "0x{:04x}".format(dl_type)
        LOG.debug("Add flow firewall log %s", str(kwargs))
        if self._deferred:
            self.int_br.add_flow(**kwargs)
        else:
            self.int_br.br.add_flow(**kwargs)

    def _delete_flows(self, **kwargs):
        ovsfw.create_reg_numbers(kwargs)
        if self._deferred:
            self.int_br.delete_flows(**kwargs)
        else:
            self.int_br.br.delete_flows(**kwargs)

    def _log_drop_packet(self, port, log_id, project_id):
        cookie = self.generate_cookie(port.id, log_const.DROP_EVENT,
                                      log_id, project_id)
        self._add_flow(
            cookie=cookie,
            table=ovs_consts.DROPPED_TRAFFIC_TABLE,
            priority=53,
            reg_port=port.ofport,
            actions='controller'
        )

    def create_rules_generator_for_port(self, port):
        for rule in port.secgroup_rules:
            yield rule

    def _create_conj_flows_log(self, remote_rule, port):
        ethertype = remote_rule['ethertype']
        direction = remote_rule['direction']
        remote_sg_id = remote_rule.get('remote_group_id')
        remote_ag_id = remote_rule.get('remote_address_group_id')
        secgroup_id = remote_rule['security_group_id']
        # we only want to log first accept packet, that means a packet with
        # ct_state=+new-est, reg_remote_group=conj_id + 1 will be logged
        flow_template = {
            'priority': REMOTE_RULE_PRIORITY,
            'dl_type': ovsfw_consts.ethertype_to_dl_type_map[ethertype],
            'reg_port': port.ofport,
            'reg_remote_group': self.conj_id_map.get_conj_id(
                secgroup_id, remote_sg_id or remote_ag_id,
                direction, ethertype) + 1,
        }
        if direction == lib_const.INGRESS_DIRECTION:
            flow_template['table'] = ovs_consts.RULES_INGRESS_TABLE
        elif direction == lib_const.EGRESS_DIRECTION:
            flow_template['table'] = ovs_consts.RULES_EGRESS_TABLE
        return [flow_template]

    def _log_accept_packet(self, port, log_id, project_id):
        cookie = self.generate_cookie(port.id, log_const.ACCEPT_EVENT,
                                      log_id, project_id)
        for rule in self.create_rules_generator_for_port(port):
            if 'remote_group_id' in rule or 'remote_address_group_id' in rule:
                flows = self._create_conj_flows_log(rule, port)
            else:
                flows = rules.create_flows_from_rule_and_port(rule, port)
            for flow in flows:
                flow['cookie'] = cookie
                self._log_accept_flow(**flow)

    def add_flows_from_rules(self, port, log_info):
        # log event can be ACCEPT or DROP or ALL(both ACCEPT and DROP)
        event = log_info['event']
        project_id = log_info['project_id']
        log_id = log_info['id']
        if event == log_const.ACCEPT_EVENT:
            self._log_accept_packet(port, log_id, project_id)
        elif event == log_const.DROP_EVENT:
            self._log_drop_packet(port, log_id, project_id)
        else:
            self._log_drop_packet(port, log_id, project_id)
            self._log_accept_packet(port, log_id, project_id)

    def _delete_accept_flows_log(self, port, log_id):
        cookie = self._get_cookie(port.id, log_const.ACCEPT_EVENT)
        if cookie:
            cookie.remove_log_obj_ref(log_id)
            if cookie.is_empty:
                self._delete_flows(
                    table=ovs_consts.ACCEPTED_INGRESS_TRAFFIC_TABLE,
                    cookie=cookie.id)
                self._delete_flows(
                    table=ovs_consts.ACCEPTED_EGRESS_TRAFFIC_TABLE,
                    cookie=cookie.id)
                self._schedule_cookie_deletion(cookie)

    def _delete_drop_flows_log(self, port, log_id):
        cookie = self._get_cookie(port.id, log_const.DROP_EVENT)
        if cookie:
            cookie.remove_log_obj_ref(log_id)
            if cookie.is_empty:
                self._delete_flows(table=ovs_consts.DROPPED_TRAFFIC_TABLE,
                                   cookie=cookie.id)
                self._schedule_cookie_deletion(cookie)

    def delete_port_flows_log(self, port, log_id):
        """Delete all flows log for given port and log_id"""
        event = port.event
        if event == log_const.ACCEPT_EVENT:
            self._delete_accept_flows_log(port, log_id)
        elif event == log_const.DROP_EVENT:
            self._delete_drop_flows_log(port, log_id)
        else:
            self._delete_accept_flows_log(port, log_id)
            self._delete_drop_flows_log(port, log_id)

    def _delete_sg_rules_flow_log(self, port, del_rules):
        cookie = self._get_cookie(port.id, log_const.ACCEPT_EVENT)
        if not cookie:
            return
        for rule in del_rules:
            if 'remote_group_id' in rule or 'remote_address_group_id' in rule:
                flows = self._create_conj_flows_log(rule, port)
            else:
                flows = rules.create_flows_from_rule_and_port(rule, port)
            for flow in flows:
                for kw in FIELDS_TO_REMOVE:
                    flow.pop(kw, None)
                flow['table'] = OVS_FW_TO_LOG_TABLES[flow['table']]
                flow['cookie'] = cookie.id
                self._delete_flows(**flow)



# Copyright 2016 Comcast
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from unittest import mock

from oslo_utils import uuidutils

from neutron.agent.l3 import agent
from neutron.agent.l3 import l3_agent_extension_api as l3_agent_api
from neutron.agent.l3 import router_info as l3router
from neutron.agent.linux import ip_lib
from neutron.conf.agent import common as config
from neutron.conf.agent.l3 import config as l3_config
from neutron.tests import base


class TestL3AgentExtensionApi(base.BaseTestCase):

    def _prepare_router_data(self, ports=None):
        self.router_id = uuidutils.generate_uuid()
        self.project_id = uuidutils.generate_uuid()
        self.conf = config.setup_conf()
        l3_config.register_l3_agent_config_opts(l3_config.OPTS, self.conf)
        ri_kwargs = {'router': {'id': self.router_id,
                                'project_id': self.project_id},
                     'agent_conf': self.conf,
                     'interface_driver': mock.ANY,
                     'use_ipv6': mock.ANY}
        ri = l3router.RouterInfo(mock.Mock(), self.router_id, **ri_kwargs)
        ri.internal_ports = ports
        return {ri.router_id: ri}, ri

    def test_get_router_hosting_port_for_router_not_in_ns(self):
        port_ids = [1, 2]
        ports = [{'id': pid} for pid in port_ids]
        router_info, ri = self._prepare_router_data(ports)

        with mock.patch.object(ip_lib,
                               'list_network_namespaces') as mock_list_netns:

            mock_list_netns.return_value = []
            api_object = l3_agent_api.L3AgentExtensionAPI(router_info, None)
            router = api_object.get_router_hosting_port(port_ids[0])

        mock_list_netns.assert_called_once_with()
        self.assertFalse(router)

    def test_get_router_hosting_port_for_router_in_ns(self):
        port_ids = [1, 2]
        ports = [{'id': pid} for pid in port_ids]
        router_info, ri = self._prepare_router_data(ports)

        with mock.patch.object(ip_lib,
                               'list_network_namespaces') as mock_list_netns:
            mock_list_netns.return_value = [ri.ns_name]
            api_object = l3_agent_api.L3AgentExtensionAPI(router_info, None)
            router = api_object.get_router_hosting_port(port_ids[0])
            self.assertEqual(ri, router)

    def test_get_routers_in_project(self):
        router_info, ri = self._prepare_router_data()

        with mock.patch.object(ip_lib,
                               'list_network_namespaces') as mock_list_netns:
            mock_list_netns.return_value = [ri.ns_name]
            api_object = l3_agent_api.L3AgentExtensionAPI(router_info, None)
            routers = api_object.get_routers_in_project(self.project_id)
            self.assertEqual([ri], routers)

    def test_is_router_in_namespace_for_in_ns(self):
        router_info, ri = self._prepare_router_data()

        with mock.patch.object(ip_lib,
                               'list_network_namespaces') as mock_list_netns:
            mock_list_netns.return_value = [ri.ns_name]
            api_object = l3_agent_api.L3AgentExtensionAPI(router_info, None)
            router_in_ns = api_object.is_router_in_namespace(ri.router_id)
            self.assertTrue(router_in_ns)

    def test_is_router_in_namespace_for_not_in_ns(self):
        router_info, ri = self._prepare_router_data()

        with mock.patch.object(ip_lib,
                               'list_network_namespaces') as mock_list_netns:
            mock_list_netns.return_value = [uuidutils.generate_uuid()]
            api_object = l3_agent_api.L3AgentExtensionAPI(router_info, None)
            router_in_ns = api_object.is_router_in_namespace(ri.router_id)
            self.assertFalse(router_in_ns)

    def test_get_router_info(self):
        router_info, ri = self._prepare_router_data()
        api_object = l3_agent_api.L3AgentExtensionAPI(router_info, None)
        self.assertEqual(ri, api_object.get_router_info(self.router_id))

    def test_get_router_info_nonexistent(self):
        router_info, ri = self._prepare_router_data()
        api_object = l3_agent_api.L3AgentExtensionAPI(router_info, None)
        self.assertIsNone(
            api_object.get_router_info(uuidutils.generate_uuid()))

    def test_register_router(self):
        router_info, ri = self._prepare_router_data()
        router_info_cls = l3router.BaseRouterInfo
        router_factory = agent.RouterFactory()
        api_object = l3_agent_api.L3AgentExtensionAPI(router_info,
                                                      router_factory)
        self.assertIsNone(
            api_object.register_router([], router_info_cls))
        self.assertIsNone(
            api_object.register_router(['ha'], router_info_cls))
        self.assertIsNone(
            api_object.register_router(['distributed'], router_info_cls))
        self.assertIsNone(
            api_object.register_router(['ha', 'distributed'], router_info_cls))



# Copyright 2015 Cloudbase Solutions.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib

import eventlet
from oslo_config import cfg
from oslo_log import log as logging

from neutron.agent.common import async_process
from neutron.agent.common import base_polling
from neutron.agent.common import ovsdb_monitor
from neutron.plugins.ml2.drivers.openvswitch.agent.common import constants

LOG = logging.getLogger(__name__)


@contextlib.contextmanager
def get_polling_manager(minimize_polling=False,
                        ovsdb_monitor_respawn_interval=(
                            constants.DEFAULT_OVSDBMON_RESPAWN),
                        bridge_names=None, ovs=None):
    if minimize_polling:
        pm = InterfacePollingMinimizer(
            ovsdb_monitor_respawn_interval=ovsdb_monitor_respawn_interval,
            bridge_names=bridge_names, ovs=ovs)
        pm.start()
    else:
        pm = base_polling.AlwaysPoll()
    try:
        yield pm
    finally:
        if minimize_polling:
            pm.stop()


class InterfacePollingMinimizer(base_polling.BasePollingManager):
    """Monitors ovsdb to determine when polling is required."""

    def __init__(
            self,
            ovsdb_monitor_respawn_interval=constants.DEFAULT_OVSDBMON_RESPAWN,
            bridge_names=None, ovs=None):

        super(InterfacePollingMinimizer, self).__init__()
        self._monitor = ovsdb_monitor.SimpleInterfaceMonitor(
            respawn_interval=ovsdb_monitor_respawn_interval,
            ovsdb_connection=cfg.CONF.OVS.ovsdb_connection,
            bridge_names=bridge_names, ovs=ovs)

    def start(self):
        self._monitor.start(block=True)

    def stop(self):
        try:
            self._monitor.stop()
        except async_process.AsyncProcessException:
            LOG.debug("InterfacePollingMinimizer was not running when stopped")

    def _is_polling_required(self):
        # Maximize the chances of update detection having a chance to
        # collect output.
        eventlet.sleep()
        return self._monitor.has_updates

    def get_events(self):
        return self._monitor.get_events()


def filter_bridge_names(br_names):
    """Bridge names to filter events received in the Interface monitor

    This method is used only in fullstack testing. Because several OVS agents
    are executed in the same host and share the same OVS switch, this filtering
    will remove events of other agents; the Interface monitor will only return
    events of Interfaces attached to Ports that belong to bridges "br_names".

    If the list is empty, no filtering is done.
    """
    return []



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib import context
from neutron_lib import fixture
from oslo_utils import uuidutils

from neutron.conf import quota as quota_conf
from neutron.db.quota import api as quota_db_api
from neutron import quota
from neutron.tests.unit.api import test_extensions
from neutron.tests.unit.db import test_db_base_plugin_v2
from neutron.tests.unit.extensions import test_l3
from neutron.tests.unit.extensions import test_securitygroup
from neutron.tests.unit.plugins.ml2 import base as ml2_base
from neutron.tests.unit.plugins.ml2 import test_plugin


class SgTestCaseWrapper(test_securitygroup.SecurityGroupDBTestCase):
    # This wrapper class enables Ml2PluginV2TestCase to correctly call the
    # setup method in SecurityGroupDBTestCase which does not accept the
    # service_plugins keyword parameter.

    def setUp(self, plugin, **kwargs):
        super(SgTestCaseWrapper, self).setUp(plugin)


class BaseTestTrackedResources(test_plugin.Ml2PluginV2TestCase,
                               SgTestCaseWrapper):

    def setUp(self):
        self.ctx = context.get_admin_context()
        self.addCleanup(self._cleanup)
        test_db_base_plugin_v2.NeutronDbPluginV2TestCase.quota_db_driver = (
            'neutron.db.quota.driver.DbQuotaDriver')
        super(BaseTestTrackedResources, self).setUp()
        self._project_id = uuidutils.generate_uuid()
        # TODO(ralonsoh): "tenant_id" reference should be removed.
        self._tenant_id = self._project_id

    @staticmethod
    def _cleanup():
        test_db_base_plugin_v2.NeutronDbPluginV2TestCase.quota_db_driver = (
            quota_conf.QUOTA_DB_DRIVER)

    def _test_init(self, resource_name):
        quota_db_api.set_quota_usage(
            self.ctx, resource_name, self._project_id)


class BaseTestEventHandler(object):

    def setUp(self):
        # Prevent noise from default security group operations
        def_sec_group_patch = mock.patch(
            'neutron.db.securitygroups_db.SecurityGroupDbMixin.'
            '_ensure_default_security_group')
        def_sec_group_patch.start()
        get_sec_group_port_patch = mock.patch(
            'neutron.db.securitygroups_db.SecurityGroupDbMixin.'
            '_get_security_groups_on_port')

        get_sec_group_port_patch.start()
        process_port_create_security_group_patch = mock.patch(
            'neutron.db.securitygroups_db.SecurityGroupDbMixin.'
            '_process_port_create_security_group')
        process_port_create_security_group_patch.start()
        handler_patch = mock.patch(
            'neutron.quota.resource.TrackedResource._db_event_handler')
        self.handler_mock = handler_patch.start()
        super(BaseTestEventHandler, self).setUp()

    def _verify_event_handler_calls(self, data, expected_call_count=1):
        if not hasattr(data, '__iter__') or isinstance(data, dict):
            data = [data]
        self.assertEqual(expected_call_count, self.handler_mock.call_count)
        call_idx = -1
        for item in data:
            if item:
                model = self.handler_mock.call_args_list[call_idx][0][-1]
                self.assertEqual(model['id'], item['id'])
                self.assertEqual(model['project_id'], item['project_id'])
            call_idx = call_idx - 1


class TestTrackedResourcesEventHandler(BaseTestEventHandler,
                                       BaseTestTrackedResources):

    def test_create_delete_network_triggers_event(self):
        self._test_init('network')
        net = self._make_network('json', 'meh', True)['network']
        self._verify_event_handler_calls(net)
        self._delete('networks', net['id'])
        self._verify_event_handler_calls(net, expected_call_count=2)

    def test_create_delete_port_triggers_event(self):
        self._test_init('port')
        net = self._make_network('json', 'meh', True)['network']
        port = self._make_port('json', net['id'])['port']
        # Expecting 2 calls - 1 for the network, 1 for the port
        self._verify_event_handler_calls(port, expected_call_count=2)
        self._delete('ports', port['id'])
        self._verify_event_handler_calls(port, expected_call_count=3)

    def test_create_delete_subnet_triggers_event(self):
        self._test_init('subnet')
        net = self._make_network('json', 'meh', True)
        subnet = self._make_subnet('json', net, '10.0.0.1',
                                   '10.0.0.0/24')['subnet']
        # Expecting 2 calls - 1 for the network, 1 for the subnet
        self._verify_event_handler_calls([subnet, net['network']],
                                         expected_call_count=2)
        self._delete('subnets', subnet['id'])
        self._verify_event_handler_calls(subnet, expected_call_count=3)

    def test_create_delete_network_with_subnet_triggers_event(self):
        self._test_init('network')
        self._test_init('subnet')
        net = self._make_network('json', 'meh', True)
        subnet = self._make_subnet('json', net, '10.0.0.1',
                                   '10.0.0.0/24')['subnet']
        # Expecting 2 calls - 1 for the network, 1 for the subnet
        self._verify_event_handler_calls([subnet, net['network']],
                                         expected_call_count=2)
        self._delete('networks', net['network']['id'])
        # Expecting 2 more calls - 1 for the network, 1 for the subnet
        self._verify_event_handler_calls([net['network'], subnet],
                                         expected_call_count=4)

    def test_create_delete_subnetpool_triggers_event(self):
        self._test_init('subnetpool')
        pool = self._make_subnetpool('json', ['10.0.0.0/8'],
                                     name='meh',
                                     tenant_id=self._project_id)['subnetpool']
        self._verify_event_handler_calls(pool)
        self._delete('subnetpools', pool['id'])
        self._verify_event_handler_calls(pool, expected_call_count=2)

    def test_create_delete_securitygroup_triggers_event(self):
        self._test_init('security_group')
        sec_group = self._make_security_group(
            'json', 'meh', 'meh',
            project_id=self._project_id)['security_group']
        # When a security group is created it also creates 2 rules, therefore
        # there will be three calls and we need to verify the first
        self._verify_event_handler_calls([None, None, sec_group],
                                         expected_call_count=3)
        self._delete('security-groups', sec_group['id'])
        # When a security group is deleted it also removes the 2 rules
        # generated upon creation
        self._verify_event_handler_calls(sec_group, expected_call_count=6)

    def test_create_delete_securitygrouprule_triggers_event(self):
        self._test_init('security_group_rule')
        sec_group = self._make_security_group(
            'json', 'meh', 'meh',
            project_id=self._project_id)['security_group']
        rule_req = self._build_security_group_rule(
            sec_group['id'], 'ingress', 'TCP', tenant_id=self._project_id)
        sec_group_rule = self._make_security_group_rule(
            'json', rule_req)['security_group_rule']
        # When a security group is created it also creates 2 rules, therefore
        # there will be four calls in total to the event handler
        self._verify_event_handler_calls(sec_group_rule, expected_call_count=4)
        self._delete('security-group-rules', sec_group_rule['id'])
        self._verify_event_handler_calls(sec_group_rule, expected_call_count=5)


class TestL3ResourcesEventHandler(BaseTestEventHandler,
                                  ml2_base.ML2TestFramework,
                                  test_l3.L3NatTestCaseMixin):

    def setUp(self):
        self.addCleanup(self._cleanup)
        test_db_base_plugin_v2.NeutronDbPluginV2TestCase.quota_db_driver = (
            'neutron.db.quota.driver.DbQuotaDriver')
        super(TestL3ResourcesEventHandler, self).setUp()
        self.useFixture(fixture.APIDefinitionFixture())
        ext_mgr = test_l3.L3TestExtensionManager()
        self.ext_api = test_extensions.setup_extensions_middleware(ext_mgr)
        quota.QUOTAS._driver = None

    @staticmethod
    def _cleanup():
        test_db_base_plugin_v2.NeutronDbPluginV2TestCase.quota_db_driver = (
            quota_conf.QUOTA_DB_DRIVER)

    def test_create_delete_floating_ip_triggers_event(self):
        net = self._make_network('json', 'meh', True)
        subnet = self._make_subnet('json', net, '14.0.0.1',
                                   '14.0.0.0/24')['subnet']
        self._set_net_external(subnet['network_id'])
        floatingip = self._make_floatingip('json', subnet['network_id'])
        internal_port = self._show(
            'ports', floatingip['floatingip']['port_id'])['ports'][0]
        # When a floatingip is created it also creates port, therefore
        # there will be four calls in total to the event handler
        self._verify_event_handler_calls(floatingip['floatingip'],
                                         expected_call_count=4)
        self._delete('floatingips', floatingip['floatingip']['id'])
        # Expecting 2 more calls - 1 for the port, 1 for the floatingip
        self._verify_event_handler_calls(
            [internal_port, floatingip['floatingip']], expected_call_count=6)


class TestTrackedResources(BaseTestTrackedResources):

    def _verify_dirty_bit(self, resource_name, expected_value=True):
        usage = quota_db_api.get_quota_usage_by_resource_and_project(
            self.ctx, resource_name, self._project_id)
        self.assertEqual(expected_value, usage.dirty)

    def test_create_delete_network_marks_dirty(self):
        self._test_init('network')
        net = self._make_network('json', 'meh', True)['network']
        self._verify_dirty_bit('network')
        # Clear the dirty bit
        quota_db_api.set_quota_usage_dirty(
            self.ctx, 'network', self._project_id, dirty=False)
        self._delete('networks', net['id'])
        self._verify_dirty_bit('network')

    def test_list_networks_clears_dirty(self):
        self._test_init('network')
        net = self._make_network('json', 'meh', True)['network']
        self.ctx.project_id = net['project_id']
        self._list('networks', neutron_context=self.ctx)
        self._verify_dirty_bit('network', expected_value=False)

    def test_create_delete_port_marks_dirty(self):
        self._test_init('port')
        net = self._make_network('json', 'meh', True)['network']
        port = self._make_port('json', net['id'])['port']
        self._verify_dirty_bit('port')
        # Clear the dirty bit
        quota_db_api.set_quota_usage_dirty(
            self.ctx, 'port', self._project_id, dirty=False)
        self._delete('ports', port['id'])
        self._verify_dirty_bit('port')

    def test_list_ports_clears_dirty(self):
        self._test_init('port')
        net = self._make_network('json', 'meh', True)['network']
        port = self._make_port('json', net['id'])['port']
        self.ctx.project_id = port['project_id']
        self._list('ports', neutron_context=self.ctx)
        self._verify_dirty_bit('port', expected_value=False)

    def test_create_delete_subnet_marks_dirty(self):
        self._test_init('subnet')
        net = self._make_network('json', 'meh', True)
        subnet = self._make_subnet('json', net, '10.0.0.1',
                                   '10.0.0.0/24')['subnet']
        self._verify_dirty_bit('subnet')
        # Clear the dirty bit
        quota_db_api.set_quota_usage_dirty(
            self.ctx, 'subnet', self._project_id, dirty=False)
        self._delete('subnets', subnet['id'])
        self._verify_dirty_bit('subnet')

    def test_create_delete_network_with_subnet_marks_dirty(self):
        self._test_init('network')
        self._test_init('subnet')
        net = self._make_network('json', 'meh', True)
        self._make_subnet('json', net, '10.0.0.1',
                          '10.0.0.0/24')['subnet']
        self._verify_dirty_bit('subnet')
        # Clear the dirty bit
        quota_db_api.set_quota_usage_dirty(
            self.ctx, 'subnet', self._project_id, dirty=False)
        self._delete('networks', net['network']['id'])
        self._verify_dirty_bit('network')
        self._verify_dirty_bit('subnet')

    def test_list_subnets_clears_dirty(self):
        self._test_init('subnet')
        net = self._make_network('json', 'meh', True)
        subnet = self._make_subnet('json', net, '10.0.0.1',
                                   '10.0.0.0/24')['subnet']
        self.ctx.project_id = subnet['project_id']
        self._list('subnets', neutron_context=self.ctx)
        self._verify_dirty_bit('subnet', expected_value=False)

    def test_create_delete_subnetpool_marks_dirty(self):
        self._test_init('subnetpool')
        pool = self._make_subnetpool('json', ['10.0.0.0/8'],
                                     name='meh',
                                     tenant_id=self._project_id)['subnetpool']
        self._verify_dirty_bit('subnetpool')
        # Clear the dirty bit
        quota_db_api.set_quota_usage_dirty(
            self.ctx, 'subnetpool', self._project_id, dirty=False)
        self._delete('subnetpools', pool['id'])
        self._verify_dirty_bit('subnetpool')

    def test_list_subnetpools_clears_dirty(self):
        self._test_init('subnetpool')
        pool = self._make_subnetpool('json', ['10.0.0.0/8'],
                                     name='meh',
                                     tenant_id=self._project_id)['subnetpool']
        self.ctx.project_id = pool['project_id']
        self._list('subnetpools', neutron_context=self.ctx)
        self._verify_dirty_bit('subnetpool', expected_value=False)

    def test_create_delete_securitygroup_marks_dirty(self):
        self._test_init('security_group')
        sec_group = self._make_security_group(
            'json', 'meh', 'meh', tenant_id=self._project_id)['security_group']
        self._verify_dirty_bit('security_group')
        # Clear the dirty bit
        quota_db_api.set_quota_usage_dirty(
            self.ctx, 'security_group', self._project_id, dirty=False)
        self._delete('security-groups', sec_group['id'])
        self._verify_dirty_bit('security_group')

    def test_list_securitygroups_clears_dirty(self):
        self._test_init('security_group')
        self._make_security_group(
            'json', 'meh', 'meh', tenant_id=self._project_id)['security_group']
        self.ctx.project_id = self._project_id
        self._list('security-groups', neutron_context=self.ctx)
        self._verify_dirty_bit('security_group', expected_value=False)

    def test_create_delete_securitygrouprule_marks_dirty(self):
        self._test_init('security_group_rule')
        sec_group = self._make_security_group(
            'json', 'meh', 'meh', tenant_id=self._project_id)['security_group']
        rule_req = self._build_security_group_rule(
            sec_group['id'], 'ingress', 'TCP', tenant_id=self._project_id)
        sec_group_rule = self._make_security_group_rule(
            'json', rule_req)['security_group_rule']
        self._verify_dirty_bit('security_group_rule')
        # Clear the dirty bit
        quota_db_api.set_quota_usage_dirty(
            self.ctx, 'security_group_rule', self._project_id, dirty=False)
        self._delete('security-group-rules', sec_group_rule['id'])
        self._verify_dirty_bit('security_group_rule')

    def test_list_securitygrouprules_clears_dirty(self):
        self._test_init('security_group_rule')
        self._make_security_group(
            'json', 'meh', 'meh', tenant_id=self._project_id)['security_group']
        # As the security group create operation also creates 2 security group
        # rules there is no need to explicitly create any rule
        self.ctx.project_id = self._project_id
        self._list('security-group-rules', neutron_context=self.ctx)
        self._verify_dirty_bit('security_group_rule', expected_value=False)



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

# Initial operations for ML2 plugin and drivers


from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'ml2_vlan_allocations',
        sa.Column('physical_network', sa.String(length=64), nullable=False),
        sa.Column('vlan_id', sa.Integer(), autoincrement=False,
                  nullable=False),
        sa.Column('allocated', sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint('physical_network', 'vlan_id'),
        sa.Index(op.f('ix_ml2_vlan_allocations_physical_network_allocated'),
                 'physical_network', 'allocated'))

    op.create_table(
        'ml2_vxlan_endpoints',
        sa.Column('ip_address', sa.String(length=64), nullable=False),
        sa.Column('udp_port', sa.Integer(), autoincrement=False,
                  nullable=False),
        sa.Column('host', sa.String(length=255), nullable=True),
        sa.UniqueConstraint('host', name='unique_ml2_vxlan_endpoints0host'),
        sa.PrimaryKeyConstraint('ip_address'))

    op.create_table(
        'ml2_gre_endpoints',
        sa.Column('ip_address', sa.String(length=64), nullable=False),
        sa.Column('host', sa.String(length=255), nullable=True),
        sa.UniqueConstraint('host', name='unique_ml2_gre_endpoints0host'),
        sa.PrimaryKeyConstraint('ip_address'))

    op.create_table(
        'ml2_vxlan_allocations',
        sa.Column('vxlan_vni', sa.Integer(), autoincrement=False,
                  nullable=False),
        sa.Column('allocated', sa.Boolean(), nullable=False,
                  server_default=sa.sql.false(), index=True),
        sa.PrimaryKeyConstraint('vxlan_vni'))

    op.create_table(
        'ml2_gre_allocations',
        sa.Column('gre_id', sa.Integer(), autoincrement=False,
                  nullable=False),
        sa.Column('allocated', sa.Boolean(), nullable=False,
                  server_default=sa.sql.false(), index=True),
        sa.PrimaryKeyConstraint('gre_id'))

    op.create_table(
        'ml2_flat_allocations',
        sa.Column('physical_network', sa.String(length=64), nullable=False),
        sa.PrimaryKeyConstraint('physical_network'))

    op.create_table(
        'ml2_network_segments',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('network_id', sa.String(length=36), nullable=False),
        sa.Column('network_type', sa.String(length=32), nullable=False),
        sa.Column('physical_network', sa.String(length=64), nullable=True),
        sa.Column('segmentation_id', sa.Integer(), nullable=True),
        sa.Column('is_dynamic', sa.Boolean(), nullable=False,
                  server_default=sa.sql.false()),
        sa.Column('segment_index', sa.Integer(), nullable=False,
                  server_default='0'),
        sa.ForeignKeyConstraint(['network_id'], ['networks.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id'))

    op.create_table(
        'ml2_port_bindings',
        sa.Column('port_id', sa.String(length=36), nullable=False),
        sa.Column('host', sa.String(length=255), nullable=False,
                  server_default=''),
        sa.Column('vif_type', sa.String(length=64), nullable=False),
        sa.Column('vnic_type', sa.String(length=64), nullable=False,
                  server_default='normal'),
        sa.Column('profile', sa.String(length=4095), nullable=False,
                  server_default=''),
        sa.Column('vif_details', sa.String(length=4095), nullable=False,
                  server_default=''),
        sa.ForeignKeyConstraint(['port_id'], ['ports.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('port_id'))

    op.create_table(
        'ml2_port_binding_levels',
        sa.Column('port_id', sa.String(length=36), nullable=False),
        sa.Column('host', sa.String(length=255), nullable=False),
        sa.Column('level', sa.Integer(), autoincrement=False, nullable=False),
        sa.Column('driver', sa.String(length=64), nullable=True),
        sa.Column('segment_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['port_id'], ['ports.id'], ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['segment_id'], ['ml2_network_segments.id'],
                                ondelete='SET NULL'),
        sa.PrimaryKeyConstraint('port_id', 'host', 'level')
    )

    op.create_table(
        'cisco_ml2_nexusport_bindings',
        sa.Column('binding_id', sa.Integer(), nullable=False),
        sa.Column('port_id', sa.String(length=255), nullable=True),
        sa.Column('vlan_id', sa.Integer(), autoincrement=False,
                  nullable=False),
        sa.Column('switch_ip', sa.String(length=255), nullable=True),
        sa.Column('instance_id', sa.String(length=255), nullable=True),
        sa.Column('vni', sa.Integer(), nullable=True),
        sa.Column('is_provider_vlan', sa.Boolean(), nullable=False,
                  server_default=sa.sql.false()),
        sa.PrimaryKeyConstraint('binding_id'),
    )

    op.create_table(
        'arista_provisioned_nets',
        sa.Column('tenant_id', sa.String(length=255), nullable=True,
                  index=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('network_id', sa.String(length=36), nullable=True),
        sa.Column('segmentation_id', sa.Integer(),
                  autoincrement=False, nullable=True),
        sa.PrimaryKeyConstraint('id'))

    op.create_table(
        'arista_provisioned_vms',
        sa.Column('tenant_id', sa.String(length=255), nullable=True,
                  index=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('vm_id', sa.String(length=255), nullable=True),
        sa.Column('host_id', sa.String(length=255), nullable=True),
        sa.Column('port_id', sa.String(length=36), nullable=True),
        sa.Column('network_id', sa.String(length=36), nullable=True),
        sa.PrimaryKeyConstraint('id'))

    op.create_table(
        'arista_provisioned_tenants',
        sa.Column('tenant_id', sa.String(length=255), nullable=True,
                  index=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.PrimaryKeyConstraint('id'))

    op.create_table(
        'ml2_nexus_vxlan_allocations',
        sa.Column('vxlan_vni', sa.Integer(), nullable=False,
                  autoincrement=False),
        sa.Column('allocated', sa.Boolean(), nullable=False,
                  server_default=sa.sql.false()),
        sa.PrimaryKeyConstraint('vxlan_vni')
    )

    op.create_table(
        'ml2_nexus_vxlan_mcast_groups',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('mcast_group', sa.String(length=64), nullable=False),
        sa.Column('associated_vni', sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(['associated_vni'],
                                ['ml2_nexus_vxlan_allocations.vxlan_vni'],
                                ondelete='CASCADE')
    )

    op.create_table(
        'cisco_ml2_nexus_nve',
        sa.Column('vni', sa.Integer(), nullable=False),
        sa.Column('switch_ip', sa.String(length=255), nullable=True),
        sa.Column('device_id', sa.String(length=255), nullable=True),
        sa.Column('mcast_group', sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint('vni', 'switch_ip', 'device_id'))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import context as n_context
from neutron_lib.db import api as db_api
from oslo_utils import uuidutils
import testtools

from neutron.common import utils
from neutron.db import models_v2
from neutron.tests.functional import base
from neutron.tests.unit import testlib_api


class TestWaitUntilTrue(base.BaseLoggingTestCase):
    def test_wait_until_true_predicate_succeeds(self):
        utils.wait_until_true(lambda: True)

    def test_wait_until_true_predicate_fails(self):
        with testtools.ExpectedException(utils.WaitTimeout):
            utils.wait_until_true(lambda: False, 2)


class _TestIsSessionActive(testlib_api.SqlTestCase):

    DRIVER = None

    def setUp(self):
        if not self.DRIVER:
            self.skipTest('No driver defined')
        super().setUp()

    def test_1(self):
        context = n_context.Context(user_id=None, tenant_id=None,
                                    is_admin=True, overwrite=False)
        self.assertFalse(utils.is_session_active(context.session))
        with db_api.CONTEXT_WRITER.using(context):
            net = models_v2.Network(id=uuidutils.generate_uuid())
            context.session.add(net)
            self.assertTrue(utils.is_session_active(context.session))

        self.assertFalse(utils.is_session_active(context.session))


class TestIsSessionActivePostgreSQL(testlib_api.PostgreSQLTestCaseMixin,
                                    _TestIsSessionActive):
    pass


class TestIsSessionActiveMySQL(testlib_api.MySQLTestCaseMixin,
                               _TestIsSessionActive):
    pass



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import copy
from unittest import mock

import fixtures
from neutron_lib import context
from neutron_lib.db import api as db_api
from neutron_lib.db import constants as db_const
from neutron_lib.exceptions import flavors as flav_exc
from neutron_lib.plugins import constants
from oslo_config import cfg
from oslo_utils import uuidutils
from webob import exc

from neutron.db.models import l3 as l3_models
from neutron.db import servicetype_db
from neutron.extensions import flavors
from neutron.objects import flavor as flavor_obj
from neutron.services.flavors import flavors_plugin
from neutron.services import provider_configuration as provconf
from neutron.tests import base
from neutron.tests.unit.api.v2 import test_base
from neutron.tests.unit.db import test_db_base_plugin_v2
from neutron.tests.unit import dummy_plugin
from neutron.tests.unit.extensions import base as extension

_uuid = uuidutils.generate_uuid
_get_path = test_base._get_path

_driver = ('neutron.tests.unit.extensions.test_flavors.'
           'DummyServiceDriver')
_provider = dummy_plugin.RESOURCE_NAME
_long_name = 'x' * (db_const.NAME_FIELD_SIZE + 1)
_long_description = 'x' * (db_const.LONG_DESCRIPTION_FIELD_SIZE + 1)


class FlavorExtensionTestCase(extension.ExtensionTestCase):

    def setUp(self):
        super(FlavorExtensionTestCase, self).setUp()
        self.setup_extension(
            'neutron.services.flavors.flavors_plugin.FlavorsPlugin',
            constants.FLAVORS, flavors.Flavors, '',
            supported_extension_aliases=['flavors'])

    def test_create_flavor(self):
        tenant_id = uuidutils.generate_uuid()
        # Use service_type FLAVORS since plugin must be loaded to validate
        data = {'flavor': {'name': 'GOLD',
                           'service_type': constants.FLAVORS,
                           'description': 'the best flavor',
                           'tenant_id': tenant_id,
                           'project_id': tenant_id,
                           'enabled': True}}

        expected = copy.deepcopy(data)
        expected['flavor']['service_profiles'] = []

        instance = self.plugin.return_value
        instance.create_flavor.return_value = expected['flavor']
        res = self.api.post(_get_path('flavors', fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt)

        instance.create_flavor.assert_called_with(mock.ANY,
                                                  flavor=expected)
        res = self.deserialize(res)
        self.assertIn('flavor', res)
        self.assertEqual(expected, res)

    def test_create_flavor_invalid_service_type(self):
        tenant_id = uuidutils.generate_uuid()
        data = {'flavor': {'name': 'GOLD',
                           'service_type': 'BROKEN',
                           'description': 'the best flavor',
                           'tenant_id': tenant_id,
                           'enabled': True}}
        self.api.post(_get_path('flavors', fmt=self.fmt),
                      self.serialize(data),
                      content_type='application/%s' % self.fmt,
                      status=exc.HTTPBadRequest.code)

    def test_create_flavor_too_long_name(self):
        tenant_id = uuidutils.generate_uuid()
        data = {'flavor': {'name': _long_name,
                           'service_type': constants.FLAVORS,
                           'description': 'the best flavor',
                           'tenant_id': tenant_id,
                           'enabled': True}}
        self.api.post(_get_path('flavors', fmt=self.fmt),
                      self.serialize(data),
                      content_type='application/%s' % self.fmt,
                      status=exc.HTTPBadRequest.code)

    def test_create_flavor_too_long_description(self):
        tenant_id = uuidutils.generate_uuid()
        data = {'flavor': {'name': _long_name,
                           'service_type': constants.FLAVORS,
                           'description': _long_description,
                           'tenant_id': tenant_id,
                           'enabled': True}}
        self.api.post(_get_path('flavors', fmt=self.fmt),
                      self.serialize(data),
                      content_type='application/%s' % self.fmt,
                      status=exc.HTTPBadRequest.code)

    def test_create_flavor_invalid_enabled(self):
        tenant_id = uuidutils.generate_uuid()
        data = {'flavor': {'name': _long_name,
                           'service_type': constants.FLAVORS,
                           'description': 'the best flavor',
                           'tenant_id': tenant_id,
                           'enabled': 'BROKEN'}}
        self.api.post(_get_path('flavors', fmt=self.fmt),
                      self.serialize(data),
                      content_type='application/%s' % self.fmt,
                      status=exc.HTTPBadRequest.code)

    def test_update_flavor(self):
        flavor_id = 'fake_id'
        data = {'flavor': {'name': 'GOLD',
                           'description': 'the best flavor',
                           'enabled': True}}
        expected = copy.copy(data)
        expected['flavor']['service_profiles'] = []

        instance = self.plugin.return_value
        instance.update_flavor.return_value = expected['flavor']
        res = self.api.put(_get_path('flavors', id=flavor_id, fmt=self.fmt),
                           self.serialize(data),
                           content_type='application/%s' % self.fmt)

        instance.update_flavor.assert_called_with(mock.ANY,
                                                  flavor_id,
                                                  flavor=expected)
        res = self.deserialize(res)
        self.assertIn('flavor', res)
        self.assertEqual(expected, res)

    def test_update_flavor_too_long_name(self):
        flavor_id = 'fake_id'
        data = {'flavor': {'name': _long_name,
                           'description': 'the best flavor',
                           'enabled': True}}
        self.api.put(_get_path('flavors', id=flavor_id, fmt=self.fmt),
                     self.serialize(data),
                     content_type='application/%s' % self.fmt,
                     status=exc.HTTPBadRequest.code)

    def test_update_flavor_too_long_description(self):
        flavor_id = 'fake_id'
        data = {'flavor': {'name': 'GOLD',
                           'description': _long_description,
                           'enabled': True}}
        self.api.put(_get_path('flavors', id=flavor_id, fmt=self.fmt),
                     self.serialize(data),
                     content_type='application/%s' % self.fmt,
                     status=exc.HTTPBadRequest.code)

    def test_update_flavor_invalid_enabled(self):
        flavor_id = 'fake_id'
        data = {'flavor': {'name': 'GOLD',
                           'description': _long_description,
                           'enabled': 'BROKEN'}}
        self.api.put(_get_path('flavors', id=flavor_id, fmt=self.fmt),
                     self.serialize(data),
                     content_type='application/%s' % self.fmt,
                     status=exc.HTTPBadRequest.code)

    def test_delete_flavor(self):
        flavor_id = 'fake_id'
        instance = self.plugin.return_value
        self.api.delete(_get_path('flavors', id=flavor_id, fmt=self.fmt),
                        content_type='application/%s' % self.fmt)

        instance.delete_flavor.assert_called_with(mock.ANY,
                                                  flavor_id)

    def test_show_flavor(self):
        flavor_id = 'fake_id'
        expected = {'flavor': {'id': flavor_id,
                               'name': 'GOLD',
                               'description': 'the best flavor',
                               'enabled': True,
                               'service_profiles': ['profile-1']}}
        instance = self.plugin.return_value
        instance.get_flavor.return_value = expected['flavor']
        res = self.api.get(_get_path('flavors', id=flavor_id, fmt=self.fmt))
        instance.get_flavor.assert_called_with(mock.ANY,
                                               flavor_id,
                                               fields=mock.ANY)
        res = self.deserialize(res)
        self.assertEqual(expected, res)

    def test_get_flavors(self):
        data = {'flavors': [{'id': 'id1',
                             'name': 'GOLD',
                             'description': 'the best flavor',
                             'enabled': True,
                             'service_profiles': ['profile-1']},
                            {'id': 'id2',
                             'name': 'GOLD',
                             'description': 'the best flavor',
                             'enabled': True,
                             'service_profiles': ['profile-2', 'profile-1']}]}
        instance = self.plugin.return_value
        instance.get_flavors.return_value = data['flavors']
        res = self.api.get(_get_path('flavors', fmt=self.fmt))
        instance.get_flavors.assert_called_with(mock.ANY,
                                                fields=mock.ANY,
                                                filters=mock.ANY)
        res = self.deserialize(res)
        self.assertEqual(data, res)

    def test_create_service_profile(self):
        tenant_id = uuidutils.generate_uuid()
        expected = {'service_profile': {'description': 'the best sp',
                                        'driver': '',
                                        'tenant_id': tenant_id,
                                        'project_id': tenant_id,
                                        'enabled': True,
                                        'metainfo': '{"data": "value"}'}}

        instance = self.plugin.return_value
        instance.create_service_profile.return_value = (
            expected['service_profile'])
        res = self.api.post(_get_path('service_profiles', fmt=self.fmt),
                            self.serialize(expected),
                            content_type='application/%s' % self.fmt)
        instance.create_service_profile.assert_called_with(
            mock.ANY,
            service_profile=expected)
        res = self.deserialize(res)
        self.assertIn('service_profile', res)
        self.assertEqual(expected, res)

    def test_create_service_profile_too_long_description(self):
        tenant_id = uuidutils.generate_uuid()
        expected = {'service_profile': {'description': _long_description,
                                        'driver': '',
                                        'tenant_id': tenant_id,
                                        'enabled': True,
                                        'metainfo': '{"data": "value"}'}}
        self.api.post(_get_path('service_profiles', fmt=self.fmt),
                      self.serialize(expected),
                      content_type='application/%s' % self.fmt,
                      status=exc.HTTPBadRequest.code)

    def test_create_service_profile_too_long_driver(self):
        tenant_id = uuidutils.generate_uuid()
        expected = {'service_profile': {'description': 'the best sp',
                                        'driver': _long_description,
                                        'tenant_id': tenant_id,
                                        'enabled': True,
                                        'metainfo': '{"data": "value"}'}}
        self.api.post(_get_path('service_profiles', fmt=self.fmt),
                      self.serialize(expected),
                      content_type='application/%s' % self.fmt,
                      status=exc.HTTPBadRequest.code)

    def test_create_service_profile_invalid_enabled(self):
        tenant_id = uuidutils.generate_uuid()
        expected = {'service_profile': {'description': 'the best sp',
                                        'driver': '',
                                        'tenant_id': tenant_id,
                                        'enabled': 'BROKEN',
                                        'metainfo': '{"data": "value"}'}}
        self.api.post(_get_path('service_profiles', fmt=self.fmt),
                      self.serialize(expected),
                      content_type='application/%s' % self.fmt,
                      status=exc.HTTPBadRequest.code)

    def test_update_service_profile(self):
        sp_id = "fake_id"
        expected = {'service_profile': {'description': 'the best sp',
                                        'enabled': False,
                                        'metainfo': '{"data1": "value3"}'}}

        instance = self.plugin.return_value
        instance.update_service_profile.return_value = (
            expected['service_profile'])
        res = self.api.put(_get_path('service_profiles',
                                     id=sp_id, fmt=self.fmt),
                           self.serialize(expected),
                           content_type='application/%s' % self.fmt)

        instance.update_service_profile.assert_called_with(
            mock.ANY,
            sp_id,
            service_profile=expected)
        res = self.deserialize(res)
        self.assertIn('service_profile', res)
        self.assertEqual(expected, res)

    def test_update_service_profile_too_long_description(self):
        sp_id = "fake_id"
        expected = {'service_profile': {'description': 'the best sp',
                                        'enabled': 'BROKEN',
                                        'metainfo': '{"data1": "value3"}'}}
        self.api.put(_get_path('service_profiles',
                               id=sp_id, fmt=self.fmt),
                     self.serialize(expected),
                     content_type='application/%s' % self.fmt,
                     status=exc.HTTPBadRequest.code)

    def test_update_service_profile_invalid_enabled(self):
        sp_id = "fake_id"
        expected = {'service_profile': {'description': 'the best sp',
                                        'enabled': 'BROKEN',
                                        'metainfo': '{"data1": "value3"}'}}
        self.api.put(_get_path('service_profiles',
                               id=sp_id, fmt=self.fmt),
                     self.serialize(expected),
                     content_type='application/%s' % self.fmt,
                     status=exc.HTTPBadRequest.code)

    def test_delete_service_profile(self):
        sp_id = 'fake_id'
        instance = self.plugin.return_value
        self.api.delete(_get_path('service_profiles', id=sp_id, fmt=self.fmt),
                        content_type='application/%s' % self.fmt)
        instance.delete_service_profile.assert_called_with(mock.ANY,
                                                           sp_id)

    def test_show_service_profile(self):
        sp_id = 'fake_id'
        expected = {'service_profile': {'id': 'id1',
                                        'driver': _driver,
                                        'description': 'desc',
                                        'metainfo': '{}',
                                        'enabled': True}}
        instance = self.plugin.return_value
        instance.get_service_profile.return_value = (
            expected['service_profile'])
        res = self.api.get(_get_path('service_profiles',
                                     id=sp_id, fmt=self.fmt))
        instance.get_service_profile.assert_called_with(mock.ANY,
                                                        sp_id,
                                                        fields=mock.ANY)
        res = self.deserialize(res)
        self.assertEqual(expected, res)

    def test_get_service_profiles(self):
        expected = {'service_profiles': [{'id': 'id1',
                                          'driver': _driver,
                                          'description': 'desc',
                                          'metainfo': '{}',
                                          'enabled': True},
                                         {'id': 'id2',
                                          'driver': _driver,
                                          'description': 'desc',
                                          'metainfo': '{}',
                                          'enabled': True}]}
        instance = self.plugin.return_value
        instance.get_service_profiles.return_value = (
            expected['service_profiles'])
        res = self.api.get(_get_path('service_profiles', fmt=self.fmt))
        instance.get_service_profiles.assert_called_with(mock.ANY,
                                                         fields=mock.ANY,
                                                         filters=mock.ANY)
        res = self.deserialize(res)
        self.assertEqual(expected, res)

    def test_associate_service_profile_with_flavor(self):
        tenant_id = uuidutils.generate_uuid()
        expected = {'service_profile': {'id': _uuid(),
                                        'tenant_id': tenant_id,
                                        'project_id': tenant_id}}
        instance = self.plugin.return_value
        instance.create_flavor_service_profile.return_value = (
            expected['service_profile'])
        res = self.api.post('/flavors/fl_id/service_profiles',
                            self.serialize(expected),
                            content_type='application/%s' % self.fmt)
        instance.create_flavor_service_profile.assert_called_with(
            mock.ANY, service_profile=expected, flavor_id='fl_id')
        res = self.deserialize(res)
        self.assertEqual(expected, res)

    def test_disassociate_service_profile_with_flavor(self):
        instance = self.plugin.return_value
        instance.delete_flavor_service_profile.return_value = None
        self.api.delete('/flavors/fl_id/service_profiles/%s' % 'fake_spid',
                        content_type='application/%s' % self.fmt)
        instance.delete_flavor_service_profile.assert_called_with(
            mock.ANY,
            'fake_spid',
            flavor_id='fl_id')

    def test_update_association_error(self):
        """Confirm that update is not permitted with user error."""
        new_id = uuidutils.generate_uuid()
        data = {'service_profile': {'id': new_id}}
        self.api.put('/flavors/fl_id/service_profiles/%s' % 'fake_spid',
                     self.serialize(data),
                     content_type='application/%s' % self.fmt,
                     status=exc.HTTPBadRequest.code)


class DummyServicePlugin(object):

    def driver_loaded(self, driver, service_profile):
        pass

    @classmethod
    def get_plugin_type(cls):
        return dummy_plugin.DUMMY_SERVICE_TYPE

    def get_plugin_description(self):
        return "Dummy service plugin, aware of flavors"


class DummyServiceDriver(object):

    @staticmethod
    def get_service_type():
        return dummy_plugin.DUMMY_SERVICE_TYPE

    def __init__(self, plugin):
        pass


class FlavorPluginTestCase(test_db_base_plugin_v2.NeutronDbPluginV2TestCase,
                           base.PluginFixture):
    def setUp(self):
        super(FlavorPluginTestCase, self).setUp()

        self.config_parse()
        cfg.CONF.set_override(
            'service_plugins',
            ['neutron.tests.unit.extensions.test_flavors.DummyServicePlugin'])

        self.useFixture(
            fixtures.MonkeyPatch('neutron.manager.NeutronManager._instance'))

        self.plugin = flavors_plugin.FlavorsPlugin()
        self.ctx = context.get_admin_context()

        providers = [DummyServiceDriver.get_service_type() +
                     ":" + _provider + ":" + _driver]
        self.service_manager = servicetype_db.ServiceTypeManager.get_instance()
        self.service_providers = mock.patch.object(
            provconf.NeutronModule, 'service_providers').start()
        self.service_providers.return_value = providers
        for provider in providers:
            self.service_manager.add_provider_configuration(
                provider.split(':')[0], provconf.ProviderConfiguration())

        db_api.CONTEXT_WRITER.get_engine()

    def _create_flavor(self, description=None):
        flavor = {'flavor': {'name': 'GOLD',
                             'service_type': dummy_plugin.DUMMY_SERVICE_TYPE,
                             'description': description or 'the best flavor',
                             'enabled': True}}
        return self.plugin.create_flavor(self.ctx, flavor), flavor

    def test_create_flavor(self):
        self._create_flavor()
        res = flavor_obj.Flavor.get_objects(self.ctx)
        self.assertEqual(1, len(res))
        self.assertEqual('GOLD', res[0]['name'])
        self.assertEqual(
            dummy_plugin.DUMMY_SERVICE_TYPE, res[0]['service_type'])

    def test_update_flavor(self):
        fl, flavor = self._create_flavor()
        flavor = {'flavor': {'name': 'Silver',
                             'enabled': False}}
        self.plugin.update_flavor(self.ctx, fl['id'], flavor)

        # don't reuse cached models from previous plugin call
        self.ctx.session.expire_all()

        res = flavor_obj.Flavor.get_object(self.ctx, id=fl['id'])
        self.assertEqual('Silver', res['name'])
        self.assertFalse(res['enabled'])

    def test_delete_flavor(self):
        fl, _ = self._create_flavor()
        self.plugin.delete_flavor(self.ctx, fl['id'])
        self.assertFalse(flavor_obj.Flavor.objects_exist(self.ctx))

    def test_show_flavor(self):
        fl, _ = self._create_flavor()
        show_fl = self.plugin.get_flavor(self.ctx, fl['id'])
        self.assertEqual(fl, show_fl)

    def test_get_flavors(self):
        fl, flavor = self._create_flavor()
        flavor['flavor']['name'] = 'SILVER'
        self.plugin.create_flavor(self.ctx, flavor)
        show_fl = self.plugin.get_flavors(self.ctx)
        self.assertEqual(2, len(show_fl))

    def _create_service_profile(self, description=None):
        data = {'service_profile':
                {'description': description or 'the best sp',
                 'driver': _driver,
                 'enabled': True,
                 'metainfo': '{"data": "value"}'}}
        sp = self.plugin.create_service_profile(self.ctx,
                                                data)
        return sp, data

    def test_create_service_profile(self):
        sp, data = self._create_service_profile()
        res = flavor_obj.ServiceProfile.get_object(self.ctx, id=sp['id'])
        self.assertIsNotNone(res)
        self.assertEqual(data['service_profile']['driver'], res.driver)
        self.assertEqual(data['service_profile']['metainfo'], res.metainfo)

    def test_create_service_profile_empty_driver(self):
        data = {'service_profile':
                {'description': 'the best sp',
                 'driver': '',
                 'enabled': True,
                 'metainfo': '{"data": "value"}'}}
        sp = self.plugin.create_service_profile(self.ctx,
                                                data)
        res = flavor_obj.ServiceProfile.get_object(self.ctx, id=sp['id'])
        self.assertIsNotNone(res)
        self.assertEqual(data['service_profile']['driver'], res.driver)
        self.assertEqual(data['service_profile']['metainfo'], res.metainfo)

    def test_create_service_profile_invalid_driver(self):
        data = {'service_profile':
                {'description': 'the best sp',
                 'driver': "Broken",
                 'enabled': True,
                 'metainfo': '{"data": "value"}'}}
        self.assertRaises(flav_exc.ServiceProfileDriverNotFound,
                          self.plugin.create_service_profile,
                          self.ctx,
                          data)

    def test_create_service_profile_invalid_empty(self):
        data = {'service_profile':
                {'description': '',
                 'driver': '',
                 'enabled': True,
                 'metainfo': ''}}
        self.assertRaises(flav_exc.ServiceProfileEmpty,
                          self.plugin.create_service_profile,
                          self.ctx,
                          data)

    def test_update_service_profile(self):
        sp, data = self._create_service_profile()
        data['service_profile']['metainfo'] = '{"data": "value1"}'
        sp = self.plugin.update_service_profile(self.ctx, sp['id'],
                                                data)

        # don't reuse cached models from previous plugin call
        self.ctx.session.expire_all()

        res = flavor_obj.ServiceProfile.get_object(self.ctx, id=sp['id'])
        self.assertEqual(data['service_profile']['metainfo'], res['metainfo'])

    def test_delete_service_profile(self):
        sp, data = self._create_service_profile()
        self.plugin.delete_service_profile(self.ctx, sp['id'])
        res = flavor_obj.ServiceProfile.get_objects(self.ctx)
        self.assertFalse(res)

    def test_show_service_profile(self):
        sp, data = self._create_service_profile()
        sp_show = self.plugin.get_service_profile(self.ctx, sp['id'])
        self.assertEqual(sp, sp_show)

    def test_get_service_profiles(self):
        self._create_service_profile()
        self._create_service_profile(description='another sp')
        self.assertEqual(2, len(self.plugin.get_service_profiles(self.ctx)))

    def test_associate_service_profile_with_flavor(self):
        sp, data = self._create_service_profile()
        fl, data = self._create_flavor()
        self.plugin.create_flavor_service_profile(
            self.ctx,
            {'service_profile': {'id': sp['id']}},
            fl['id'])
        binding = flavor_obj.FlavorServiceProfileBinding.get_objects(
            self.ctx)[0]
        self.assertEqual(fl['id'], binding['flavor_id'])
        self.assertEqual(sp['id'], binding['service_profile_id'])

        # don't reuse cached models from previous plugin call
        self.ctx.session.expire_all()

        res = self.plugin.get_flavor(self.ctx, fl['id'])
        self.assertEqual(1, len(res['service_profiles']))
        self.assertEqual(sp['id'], res['service_profiles'][0])

        res = self.plugin.get_service_profile(self.ctx, sp['id'])
        self.assertEqual(1, len(res['flavors']))
        self.assertEqual(fl['id'], res['flavors'][0])

    def test_autodelete_flavor_associations(self):
        sp, data = self._create_service_profile()
        fl, data = self._create_flavor()
        self.plugin.create_flavor_service_profile(
            self.ctx,
            {'service_profile': {'id': sp['id']}},
            fl['id'])
        self.plugin.delete_flavor(self.ctx, fl['id'])
        self.assertFalse(
            flavor_obj.FlavorServiceProfileBinding.objects_exist(self.ctx))

    def test_associate_service_profile_with_flavor_exists(self):
        sp, data = self._create_service_profile()
        fl, data = self._create_flavor()
        self.plugin.create_flavor_service_profile(
            self.ctx,
            {'service_profile': {'id': sp['id']}},
            fl['id'])
        self.assertRaises(flav_exc.FlavorServiceProfileBindingExists,
                          self.plugin.create_flavor_service_profile,
                          self.ctx,
                          {'service_profile': {'id': sp['id']}},
                          fl['id'])

    def test_disassociate_service_profile_with_flavor(self):
        sp, data = self._create_service_profile()
        fl, data = self._create_flavor()
        self.plugin.create_flavor_service_profile(
            self.ctx,
            {'service_profile': {'id': sp['id']}},
            fl['id'])
        self.plugin.delete_flavor_service_profile(
            self.ctx, sp['id'], fl['id'])

        self.assertFalse(
            flavor_obj.FlavorServiceProfileBinding.objects_exist(self.ctx))

        self.assertRaises(
            flav_exc.FlavorServiceProfileBindingNotFound,
            self.plugin.delete_flavor_service_profile,
            self.ctx, sp['id'], fl['id'])

    def test_delete_service_profile_in_use(self):
        sp, data = self._create_service_profile()
        fl, data = self._create_flavor()
        self.plugin.create_flavor_service_profile(
            self.ctx,
            {'service_profile': {'id': sp['id']}},
            fl['id'])
        self.assertRaises(
            flav_exc.ServiceProfileInUse,
            self.plugin.delete_service_profile,
            self.ctx,
            sp['id'])

    def test_delete_flavor_in_use(self):
        # make use of router since it has a flavor id
        fl, data = self._create_flavor()
        with db_api.CONTEXT_WRITER.using(self.ctx):
            self.ctx.session.add(l3_models.Router(flavor_id=fl['id']))
        self.assertRaises(
            flav_exc.FlavorInUse,
            self.plugin.delete_flavor,
            self.ctx,
            fl['id'])

    def test_get_flavor_next_provider_no_binding(self):
        fl, data = self._create_flavor()
        self.assertRaises(
            flav_exc.FlavorServiceProfileBindingNotFound,
            self.plugin.get_flavor_next_provider,
            self.ctx,
            fl['id'])

    def test_get_flavor_next_provider_disabled(self):
        data = {'service_profile':
                {'description': 'the best sp',
                 'driver': _driver,
                 'enabled': False,
                 'metainfo': '{"data": "value"}'}}
        sp = self.plugin.create_service_profile(self.ctx,
                                                data)
        fl, data = self._create_flavor()
        self.plugin.create_flavor_service_profile(
            self.ctx,
            {'service_profile': {'id': sp['id']}},
            fl['id'])
        self.assertRaises(
            flav_exc.ServiceProfileDisabled,
            self.plugin.get_flavor_next_provider,
            self.ctx,
            fl['id'])

    def test_get_flavor_next_provider_no_driver(self):
        data = {'service_profile':
                {'description': 'the best sp',
                 'driver': '',
                 'enabled': True,
                 'metainfo': '{"data": "value"}'}}
        sp = self.plugin.create_service_profile(self.ctx,
                                                data)
        fl, data = self._create_flavor()
        self.plugin.create_flavor_service_profile(
            self.ctx,
            {'service_profile': {'id': sp['id']}},
            fl['id'])
        self.assertRaises(
            flav_exc.ServiceProfileDriverNotFound,
            self.plugin.get_flavor_next_provider,
            self.ctx,
            fl['id'])

    def test_get_flavor_next_provider(self):
        sp, data = self._create_service_profile()
        fl, data = self._create_flavor()
        self.plugin.create_flavor_service_profile(
            self.ctx,
            {'service_profile': {'id': sp['id']}},
            fl['id'])
        providers = self.plugin.get_flavor_next_provider(
            self.ctx,
            fl['id'])
        self.assertEqual(_provider, providers[0].get('provider', None))



#!/usr/bin/env python3

# Copyright 2020 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re
import sys

file_names = set()


def parse_input(input_file):
    global file_names

    while True:
        line_buffer = input_file.readline()
        if not line_buffer:
            break
        line_match = re.search(r"^\s*---\s+([^\s@]+)[\s@]+", line_buffer)
        if not line_match:
            line_match = re.search(r"^\s*\+\+\+\s+([^\s@]+)[\s@]+",
                                   line_buffer)
        if line_match:
            curr_file_name = line_match.group(1)

            # trim off 'a/' and 'b/' that you will normally see in git output
            #
            if len(curr_file_name) > 2 and curr_file_name[1] == '/' and (
                    curr_file_name[0] == 'a' or curr_file_name[0] == 'b'):
                curr_file_name = curr_file_name[2:]

            file_names.add(curr_file_name)


def prune_unwanted_names():
    global file_names

    unwanted_names = set(['/dev/null'])

    for curr_file_name in file_names:
        # ignore files that end in '.orig' as long as non-.orig exists
        line_match = re.search(r"^(.+)\.[oO][Rr][iI][gG]$", curr_file_name)
        if line_match and line_match.group(1) in file_names:
            unwanted_names.add(curr_file_name)
            continue

    file_names -= unwanted_names


def print_file_names():
    for name in sorted(file_names):
        print(name)


if __name__ == '__main__':
    if len(sys.argv) == 1:
        parse_input(sys.stdin)
    else:
        for curr_input_name in sys.argv[1:]:
            try:
                with open(curr_input_name, 'r') as curr_input_file:
                    parse_input(curr_input_file)
            except IOError as e_str:
                sys.stderr.write(
                    "Cannot open {}: {}\n".format(curr_input_name, e_str))
                sys.exit(255)

    prune_unwanted_names()
    print_file_names()



# Copyright 2013 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from itertools import chain

from oslo_log import log as logging

from neutron.objects import servicetype as servicetype_obj
from neutron.services import provider_configuration as pconf

LOG = logging.getLogger(__name__)


class ServiceTypeManager(object):
    """Manage service type objects in Neutron."""

    _instance = None

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def __init__(self):
        self.config = {}

    def add_provider_configuration(self, service_type, configuration):
        """Add or update the provider configuration for the service type."""
        LOG.debug('Adding provider configuration for service %s', service_type)
        self.config.update({service_type: configuration})

    def get_service_providers(self, context, filters=None, fields=None):
        if filters and 'service_type' in filters:
            return list(chain.from_iterable(
                self.config[svc_type].get_service_providers(filters, fields)
                for svc_type in filters['service_type']
                if svc_type in self.config)
            )
        return list(
            chain.from_iterable(
                self.config[p].get_service_providers(filters, fields)
                for p in self.config)
        )

    def get_default_service_provider(self, context, service_type):
        """Return the default provider for a given service type."""
        filters = {'service_type': [service_type],
                   'default': [True]}
        providers = self.get_service_providers(context, filters=filters)
        # By construction we expect at most a single item in provider
        if not providers:
            raise pconf.DefaultServiceProviderNotFound(
                service_type=service_type
            )
        return providers[0]

    def get_provider_names_by_resource_ids(self, context, resource_ids):
        objs = servicetype_obj.ProviderResourceAssociation.get_objects(
            context, resource_id=resource_ids)
        return {rec.resource_id: rec.provider_name for rec in objs}

    def add_resource_association(self, context, service_type, provider_name,
                                 resource_id, expire_session=True):
        r = self.get_service_providers(context,
                                       filters={'service_type': [service_type],
                                                'name': [provider_name]})
        if not r:
            raise pconf.ServiceProviderNotFound(provider=provider_name,
                                                service_type=service_type)

        # we don't actually need service type for association.
        # resource_id is unique and belongs to specific service
        # which knows its type
        servicetype_obj.ProviderResourceAssociation(
            context, provider_name=provider_name,
            resource_id=resource_id).create()
        # NOTE(blogan): the ProviderResourceAssociation relationship will not
        # be populated if a resource was created before this.  The expire_all
        # will force the session to go retrieve the new data when that
        # resource will be read again.  It has been suggested that we can
        # crawl through everything in the mapper to find the resource with
        # the ID that matches resource_id and expire that one, but we can
        # just start with this.
        # NOTE(ralonsoh): to be removed once the new engine facade is fully
        # implanted in Neutron.
        if expire_session:
            context.session.expire_all()

    def del_resource_associations(self, context, resource_ids):
        if not resource_ids:
            return
        servicetype_obj.ProviderResourceAssociation.delete_objects(
            context, resource_id=resource_ids)



# Copyright (C) 2013 eNovance SAS <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys

from neutron_lib.agent import topics
from neutron_lib import constants
from neutron_lib import context
from neutron_lib import rpc as n_rpc
from neutron_lib.utils import runtime
from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging
from oslo_service import loopingcall
from oslo_service import periodic_task
from oslo_service import service
from oslo_utils import timeutils

from neutron._i18n import _
from neutron.agent import rpc as agent_rpc
from neutron.common import config as common_config
from neutron.conf.agent import common as config
from neutron.conf.services import metering_agent
from neutron import manager
from neutron import service as neutron_service

from neutron.services.metering.drivers import abstract_driver as driver
from neutron.services.metering.drivers import utils as driverutils

LOG = logging.getLogger(__name__)


class MeteringPluginRpc(object):

    def __init__(self, host):
        # NOTE(yamamoto): super.__init__() call here is not only for
        # aesthetics.  Because of multiple inheritances in MeteringAgent,
        # it's actually necessary to initialize parent classes of
        # manager.Manager correctly.
        super(MeteringPluginRpc, self).__init__(host)
        target = oslo_messaging.Target(topic=topics.METERING_PLUGIN,
                                       version='1.0')
        self.client = n_rpc.get_client(target)

    def _get_sync_data_metering(self, context):
        try:
            cctxt = self.client.prepare()
            return cctxt.call(context, 'get_sync_data_metering',
                              host=self.host)
        except Exception:
            LOG.exception("Failed synchronizing routers")


class MeteringAgent(MeteringPluginRpc, manager.Manager):

    def __init__(self, host, conf=None):
        self.conf = conf or cfg.CONF
        self._load_drivers()
        self.context = context.get_admin_context_without_session()
        self.metering_loop = loopingcall.FixedIntervalLoopingCall(
            self._metering_loop
        )
        measure_interval = self.conf.measure_interval
        self.last_report = 0
        self.metering_loop.start(interval=measure_interval)
        self.host = host

        self.label_project_id = {}
        self.routers = {}
        self.metering_infos = {}
        self.metering_labels = {}
        super(MeteringAgent, self).__init__(host=host)

    def _load_drivers(self):
        """Loads plugin-driver from configuration."""
        LOG.info("Loading Metering driver %s", self.conf.driver)
        if not self.conf.driver:
            raise SystemExit(_('A metering driver must be specified'))
        self.metering_driver = driverutils.load_metering_driver(self,
                                                                self.conf)

    def _metering_notification(self):
        for key, info in self.metering_infos.items():
            data = self.create_notification_message_data(info, key)

            traffic_meter_event = 'l3.meter'

            granularity = info.get('traffic-counter-granularity')
            if granularity:
                traffic_meter_event = 'l3.meter.%s' % granularity

            LOG.debug("Send metering report [%s] via event [%s].",
                      data, traffic_meter_event)

            notifier = n_rpc.get_notifier('metering')
            notifier.info(self.context, traffic_meter_event, data)

            info['pkts'] = 0
            info['bytes'] = 0
            info['time'] = 0

    def create_notification_message_data(self, info, key):
        data = {'pkts': info['pkts'],
                'bytes': info['bytes'],
                'time': info['time'],
                'first_update': info['first_update'],
                'last_update': info['last_update'],
                'host': self.host}

        if self.conf.granular_traffic_data:
            data['resource_id'] = key
            self.set_project_id_for_granular_traffic_data(data, key)
        else:
            data['label_id'] = key
            data['project_id'] = self.label_project_id.get(key)

        LOG.debug("Metering notification created [%s] with info data [%s], "
                  "key[%s], and metering_labels configured [%s]. ", data, info,
                  key, self.metering_labels)
        return data

    def set_project_id_for_granular_traffic_data(self, data, key):
        if driver.BASE_LABEL_TRAFFIC_COUNTER_KEY in key:
            other_ids, actual_label_id = key.split(
                driver.BASE_LABEL_TRAFFIC_COUNTER_KEY)

            is_label_shared = None
            label_name = actual_label_id

            metering_label = self.metering_labels.get(actual_label_id)
            if metering_label:
                is_label_shared = metering_label['shared']
                label_name = metering_label['name']

            data['label_id'] = actual_label_id
            data['label_name'] = label_name
            data['label_shared'] = is_label_shared

            if is_label_shared:
                self.configure_project_id_shared_labels(data, other_ids[:-1])
            else:
                data['project_id'] = self.label_project_id.get(actual_label_id)
        elif driver.BASE_PROJECT_TRAFFIC_COUNTER_KEY in key:
            data['project_id'] = key.split(
                driver.BASE_PROJECT_TRAFFIC_COUNTER_KEY)[1]
        elif driver.BASE_ROUTER_TRAFFIC_COUNTER_KEY in key:
            router_id = key.split(driver.BASE_ROUTER_TRAFFIC_COUNTER_KEY)[1]
            data['router_id'] = router_id
            self.configure_project_id_based_on_router(data, router_id)
        else:
            raise Exception(_("Unexpected key [%s] format.") % key)

    def configure_project_id_shared_labels(self, data, key):
        if driver.BASE_PROJECT_TRAFFIC_COUNTER_KEY in key:
            project_id = key.split(driver.BASE_PROJECT_TRAFFIC_COUNTER_KEY)[1]

            data['project_id'] = project_id
        elif driver.BASE_ROUTER_TRAFFIC_COUNTER_KEY in key:
            router_id = key.split(driver.BASE_ROUTER_TRAFFIC_COUNTER_KEY)[1]

            data['router_id'] = router_id
            self.configure_project_id_based_on_router(data, router_id)
        else:
            data['project_id'] = 'all'

    def configure_project_id_based_on_router(self, data, router_id):
        if router_id in self.routers:
            router = self.routers[router_id]
            data['project_id'] = router['project_id']
        else:
            LOG.warning("Could not find router with ID [%s].", router_id)

    def _purge_metering_info(self):
        deadline_timestamp = timeutils.utcnow_ts() - self.conf.report_interval
        expired_metering_info_key = [
            key for key, info in self.metering_infos.items()
            if info['last_update'] < deadline_timestamp]

        for key in expired_metering_info_key:
            del self.metering_infos[key]

    def _add_metering_info(self, key, traffic_counter):
        granularity = traffic_counter.get('traffic-counter-granularity')

        ts = timeutils.utcnow_ts()
        info = self.metering_infos.get(
            key, {'bytes': 0, 'traffic-counter-granularity': granularity,
                  'pkts': 0, 'time': 0, 'first_update': ts, 'last_update': ts})

        info['bytes'] += traffic_counter['bytes']
        info['pkts'] += traffic_counter['pkts']
        info['time'] += ts - info['last_update']
        info['last_update'] = ts

        self.metering_infos[key] = info

        return info

    def _add_metering_infos(self):
        self.label_project_id = {}
        for router in self.routers.values():
            project_id = router['project_id']
            labels = router.get(constants.METERING_LABEL_KEY, [])
            for label in labels:
                label_id = label['id']
                self.label_project_id[label_id] = project_id

        LOG.debug("Retrieving traffic counters for routers [%s].",
                  self.routers)
        traffic_counters = self._get_traffic_counters(self.context,
                                                      self.routers.values())
        LOG.debug("Traffic counters [%s] retrieved for routers [%s].",
                  traffic_counters, self.routers)
        if not traffic_counters:
            return

        for key, traffic_counter in traffic_counters.items():
            self._add_metering_info(key, traffic_counter)

    def _metering_loop(self):
        self._sync_router_namespaces(self.context, self.routers.values())
        self._add_metering_infos()

        ts = timeutils.utcnow_ts()
        delta = ts - self.last_report

        report_interval = self.conf.report_interval
        if delta >= report_interval:
            self._metering_notification()
            self._purge_metering_info()
            self.last_report = ts

    @runtime.synchronized('metering-agent')
    def _invoke_driver(self, context, meterings, func_name):
        try:
            return getattr(self.metering_driver, func_name)(context, meterings)
        except AttributeError:
            LOG.exception("Driver %(driver)s does not implement %(func)s",
                          {'driver': self.conf.driver,
                           'func': func_name})
        except RuntimeError:
            LOG.exception("Driver %(driver)s:%(func)s runtime error",
                          {'driver': self.conf.driver,
                           'func': func_name})

    @periodic_task.periodic_task(run_immediately=True)
    def _sync_routers_task(self, context):
        routers = self._get_sync_data_metering(self.context)

        routers_on_agent = set(self.routers.keys())
        routers_on_server = set(
            [router['id'] for router in routers] if routers else [])
        for router_id in routers_on_agent - routers_on_server:
            del self.routers[router_id]
            self._invoke_driver(context, router_id, 'remove_router')

        if not routers:
            return
        self._update_routers(context, routers)

    def router_deleted(self, context, router_id):
        self._add_metering_infos()

        if router_id in self.routers:
            del self.routers[router_id]

        return self._invoke_driver(context, router_id,
                                   'remove_router')

    def routers_updated(self, context, routers=None):
        if not routers:
            routers = self._get_sync_data_metering(self.context)
        if not routers:
            return
        self._update_routers(context, routers)

    def _update_routers(self, context, routers):
        for router in routers:
            self.routers[router['id']] = router

            self.store_metering_labels(router)

        return self._invoke_driver(context, routers,
                                   'update_routers')

    def _get_traffic_counters(self, context, routers):
        LOG.debug("Get router traffic counters")
        return self._invoke_driver(context, routers, 'get_traffic_counters')

    def _sync_router_namespaces(self, context, routers):
        LOG.debug("Sync router namespaces")
        return self._invoke_driver(context, routers, 'sync_router_namespaces')

    def add_metering_label_rule(self, context, routers):
        return self._invoke_driver(context, routers,
                                   'add_metering_label_rule')

    def remove_metering_label_rule(self, context, routers):
        return self._invoke_driver(context, routers,
                                   'remove_metering_label_rule')

    def update_metering_label_rules(self, context, routers):
        LOG.debug("Update metering rules from agent")
        return self._invoke_driver(context, routers,
                                   'update_metering_label_rules')

    def add_metering_label(self, context, routers):
        LOG.debug("Creating a metering label from agent with parameters ["
                  "%s].", routers)
        for router in routers:
            self.store_metering_labels(router)

        return self._invoke_driver(context, routers,
                                   'add_metering_label')

    def store_metering_labels(self, router):
        labels = router[constants.METERING_LABEL_KEY]
        for label in labels:
            self.metering_labels[label['id']] = label

    def remove_metering_label(self, context, routers):
        self._add_metering_infos()
        LOG.debug("Delete a metering label from agent with parameters ["
                  "%s].", routers)

        for router in routers:
            labels = router[constants.METERING_LABEL_KEY]
            for label in labels:
                if label['id'] in self.metering_labels.keys():
                    del self.metering_labels[label['id']]

        return self._invoke_driver(context, routers,
                                   'remove_metering_label')


class MeteringAgentWithStateReport(MeteringAgent):

    def __init__(self, host, conf=None):
        super(MeteringAgentWithStateReport, self).__init__(host=host,
                                                           conf=conf)
        self.state_rpc = agent_rpc.PluginReportStateAPI(topics.REPORTS)
        self.failed_report_state = False
        self.agent_state = {
            'binary': constants.AGENT_PROCESS_METERING,
            'host': host,
            'topic': topics.METERING_AGENT,
            'configurations': {
                'metering_driver': self.conf.driver,
                'measure_interval':
                self.conf.measure_interval,
                'report_interval': self.conf.report_interval
            },
            'start_flag': True,
            'agent_type': constants.AGENT_TYPE_METERING}
        report_interval = cfg.CONF.AGENT.report_interval
        self.use_call = True
        if report_interval:
            self.heartbeat = loopingcall.FixedIntervalLoopingCall(
                self._report_state)
            self.heartbeat.start(interval=report_interval)

    def _report_state(self):
        try:
            self.state_rpc.report_state(self.context, self.agent_state,
                                        self.use_call)
            self.agent_state.pop('start_flag', None)
            self.use_call = False
        except AttributeError:
            # This means the server does not support report_state
            LOG.warning("Neutron server does not support state report. "
                        "State report for this agent will be disabled.")
            self.heartbeat.stop()
        except Exception:
            self.failed_report_state = True
            LOG.exception("Failed reporting state!")
            return
        if self.failed_report_state:
            self.failed_report_state = False
            LOG.info("Successfully reported state after a previous failure.")

    def agent_updated(self, context, payload):
        LOG.info("agent_updated by server side %s!", payload)


def main():
    conf = cfg.CONF
    metering_agent.register_metering_agent_opts()
    config.register_agent_state_opts_helper(conf)
    common_config.init(sys.argv[1:])
    config.setup_logging()
    config.setup_privsep()
    server = neutron_service.Service.create(
        binary=constants.AGENT_PROCESS_METERING,
        topic=topics.METERING_AGENT,
        report_interval=cfg.CONF.AGENT.report_interval,
        manager='neutron.services.metering.agents.'
                'metering_agent.MeteringAgentWithStateReport')
    service.launch(cfg.CONF, server, restart_method='mutate').wait()



# Copyright (C) 2013 eNovance SAS <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import helpers as log_helpers

from neutron.services.metering.drivers import abstract_driver


class NoopMeteringDriver(abstract_driver.MeteringAbstractDriver):

    @log_helpers.log_method_call
    def update_routers(self, context, routers):
        pass

    @log_helpers.log_method_call
    def remove_router(self, context, router_id):
        pass

    @log_helpers.log_method_call
    def update_metering_label_rules(self, context, routers):
        pass

    @log_helpers.log_method_call
    def add_metering_label_rule(self, context, routers):
        pass

    @log_helpers.log_method_call
    def remove_metering_label_rule(self, context, routers):
        pass

    @log_helpers.log_method_call
    def add_metering_label(self, context, routers):
        pass

    @log_helpers.log_method_call
    def remove_metering_label(self, context, routers):
        pass

    @log_helpers.log_method_call
    def get_traffic_counters(self, context, routers):
        pass

    @log_helpers.log_method_call
    def sync_router_namespaces(self, context, routers):
        pass



# Copyright (c) 2021 Red Hat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import context as n_context
from neutron_lib.db import api as db_api
from neutron_lib import exceptions
from oslo_db import exception as db_exc
from oslo_log import log

from neutron.common import utils
from neutron.db.quota import api as quota_api
from neutron.db.quota import driver as quota_driver
from neutron import worker as neutron_worker


LOG = log.getLogger(__name__)


class DbQuotaNoLockDriver(quota_driver.DbQuotaDriver):
    """Driver to enforce quotas and retrieve quota information

    This driver does not use a (resource, project_id) lock but relays on the
    simplicity of the calculation method, that is executed in a single database
    transaction. The method (1) counts the number of created resources and (2)
    the number of resource reservations. If the requested number of resources
    do not exceeds the quota, a new reservation register is created.

    This calculation method does not guarantee the quota enforcement if, for
    example, the database isolation level is read committed; two transactions
    can count the same number of resources and reservations, committing both
    a new reservation exceeding the quota. But the goal of this reservation
    method is to be fast enough to avoid the concurrency when counting the
    resources while not blocking concurrent API operations.
    """
    @staticmethod
    @utils.skip_exceptions(db_exc.DBError)
    def _remove_expired_reservations():
        """Remove expired reservations from all projects

        Any DB exception will be catch and dismissed. This operation can have
        been successfully executed by another concurrent request. There is no
        need to fail or retry it.
        """
        context = n_context.get_admin_context()
        quota_api.remove_expired_reservations(
            context, timeout=quota_api.RESERVATION_EXPIRATION_TIMEOUT)

    @db_api.retry_if_session_inactive()
    def make_reservation(self, context, project_id, resources, deltas, plugin):
        resources_over_limit = []
        with db_api.CONTEXT_WRITER.using(context):
            # Filter out unlimited resources.
            limits = self.get_project_quotas(context, resources, project_id)
            unlimited_resources = set([resource for (resource, limit) in
                                       limits.items() if limit < 0])
            requested_resources = (set(deltas.keys()) - unlimited_resources)

            # Count the number of (1) used and (2) reserved resources for this
            # project_id. If any resource limit is exceeded, raise exception.
            for resource_name in requested_resources:
                used_and_reserved = self.get_resource_usage(
                    context, project_id, resources, resource_name)
                resource_num = deltas[resource_name]
                if limits[resource_name] < (used_and_reserved + resource_num):
                    resources_over_limit.append(resource_name)

            if resources_over_limit:
                raise exceptions.OverQuota(overs=sorted(resources_over_limit))

            return quota_api.create_reservation(context, project_id, deltas)

    def cancel_reservation(self, context, reservation_id):
        quota_api.remove_reservation(context, reservation_id, set_dirty=False)

    @staticmethod
    def get_resource_usage(context, project_id, resources, resource_name):
        tracked_resource = resources.get(resource_name)
        if not tracked_resource:
            return
        return tracked_resource.count(context, None, project_id,
                                      count_db_registers=True)

    @staticmethod
    def get_resource_count(context, project_id, tracked_resource):
        return tracked_resource.count_db_registers(context, project_id)

    @staticmethod
    def get_workers():
        interval = quota_api.RESERVATION_EXPIRATION_TIMEOUT
        method = DbQuotaNoLockDriver._remove_expired_reservations
        return [neutron_worker.PeriodicWorker(method, interval, interval)]



# Copyright 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from alembic import op
from oslo_utils import uuidutils
import sqlalchemy as sa

"""network_rbac_external

Revision ID: 5ffceebfada
Revises: c6c112992c9
Create Date: 2015-06-14 13:12:04.012457

"""

# revision identifiers, used by Alembic.
revision = '5ffceebfada'
down_revision = 'c6c112992c9'
depends_on = ()


# A simple model of the external network table with only the fields needed for
# the migration.
external = sa.Table('externalnetworks', sa.MetaData(),
                    sa.Column('network_id', sa.String(length=36),
                              nullable=False))

network = sa.Table('networks', sa.MetaData(),
                   sa.Column('id', sa.String(length=36), nullable=False),
                   sa.Column('tenant_id', sa.String(length=255)))

networkrbacs = sa.Table(
    'networkrbacs', sa.MetaData(),
    sa.Column('id', sa.String(length=36), nullable=False),
    sa.Column('object_id', sa.String(length=36), nullable=False),
    sa.Column('tenant_id', sa.String(length=255), nullable=True,
              index=True),
    sa.Column('target_tenant', sa.String(length=255),
              nullable=False),
    sa.Column('action', sa.String(length=255), nullable=False))


def upgrade():
    op.bulk_insert(networkrbacs, get_values())


def get_values():
    session = sa.orm.Session(bind=op.get_bind())
    values = []
    net_to_tenant_id = {}
    for row in session.query(network).all():
        net_to_tenant_id[row[0]] = row[1]
    for row in session.query(external).all():
        values.append({'id': uuidutils.generate_uuid(), 'object_id': row[0],
                       'tenant_id': net_to_tenant_id[row[0]],
                       'target_tenant': '*', 'action': 'access_as_external'})
    # this commit appears to be necessary to allow further operations
    session.commit()
    return values



# Copyright 2018 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

import netaddr
from neutron_lib.agent import l3_extension
from neutron_lib import constants
from neutron_lib import rpc as n_rpc
from oslo_concurrency import lockutils
from oslo_log import log as logging

from neutron.agent.linux import ip_lib
from neutron.api.rpc.callbacks.consumer import registry
from neutron.api.rpc.callbacks import events
from neutron.api.rpc.callbacks import resources
from neutron.api.rpc.handlers import resources_rpc
from neutron.common import coordination


LOG = logging.getLogger(__name__)
DEFAULT_PORT_FORWARDING_CHAIN = 'fip-pf'
PORT_FORWARDING_PREFIX = 'fip_portforwarding-'
PORT_FORWARDING_CHAIN_PREFIX = 'pf-'


class RouterFipPortForwardingMapping(object):
    def __init__(self):
        self.managed_port_forwardings = {}
        """
        fip_port_forwarding = {
            fip_id_1: set(pf_id1, pf_id2),
            fip_id_2: set(pf_id3, pf_id4)
        }
        """
        self.fip_port_forwarding = collections.defaultdict(set)
        """
        router_fip_mapping = {
            router_id_1: set(fip_id_1, fip_id_2),
            router_id_2: set(fip_id_3, fip_id_4)
        }
        """
        self.router_fip_mapping = collections.defaultdict(set)

    @lockutils.synchronized('port-forwarding-cache')
    def set_port_forwardings(self, port_forwardings):
        for port_forwarding in port_forwardings:
            self._set_fip_port_forwarding(port_forwarding.floatingip_id,
                                          port_forwarding,
                                          port_forwarding.router_id)

    @lockutils.synchronized('port-forwarding-cache')
    def update_port_forwardings(self, port_forwardings):
        for port_forwarding in port_forwardings:
            self.managed_port_forwardings[port_forwarding.id] = port_forwarding

    @lockutils.synchronized('port-forwarding-cache')
    def del_port_forwardings(self, port_forwardings):
        for port_forwarding in port_forwardings:
            if not self.managed_port_forwardings.get(port_forwarding.id):
                continue
            self.managed_port_forwardings.pop(port_forwarding.id)
            self.fip_port_forwarding[port_forwarding.floatingip_id].discard(
                port_forwarding.id)
            if not self.fip_port_forwarding[port_forwarding.floatingip_id]:
                self.fip_port_forwarding.pop(port_forwarding.floatingip_id)
                self.router_fip_mapping[port_forwarding.router_id].discard(
                    port_forwarding.floatingip_id)
                if not self.router_fip_mapping[port_forwarding.router_id]:
                    del self.router_fip_mapping[port_forwarding.router_id]

    def _set_fip_port_forwarding(self, fip_id, pf, router_id):
        self.router_fip_mapping[router_id].add(fip_id)
        self.fip_port_forwarding[fip_id].add(pf.id)
        self.managed_port_forwardings[pf.id] = pf

    @lockutils.synchronized('port-forwarding-cache')
    def clear_by_fip(self, fip_id, router_id):
        self.router_fip_mapping[router_id].discard(fip_id)
        if len(self.router_fip_mapping[router_id]) == 0:
            del self.router_fip_mapping[router_id]
        for pf_id in self.fip_port_forwarding[fip_id]:
            del self.managed_port_forwardings[pf_id]
        del self.fip_port_forwarding[fip_id]

    @lockutils.synchronized('port-forwarding-cache')
    def check_port_forwarding_changes(self, new_pf):
        old_pf = self.managed_port_forwardings.get(new_pf.id)
        return old_pf != new_pf

    @lockutils.synchronized('port-forwarding-cache')
    def clean_port_forwardings_by_router_id(self, router_id):
        router_fips = self.router_fip_mapping.pop(router_id, [])
        for fip_id in router_fips:
            pf_ids = self.fip_port_forwarding.pop(fip_id, [])
            for pf_id in pf_ids:
                self.managed_port_forwardings.pop(pf_id, None)


class PortForwardingAgentExtension(l3_extension.L3AgentExtension):
    SUPPORTED_RESOURCE_TYPES = [resources.PORTFORWARDING]

    def initialize(self, connection, driver_type):
        self.resource_rpc = resources_rpc.ResourcesPullRpcApi()
        self._register_rpc_consumers()
        self.mapping = RouterFipPortForwardingMapping()

    def _register_rpc_consumers(self):
        registry.register(self._handle_notification,
                          resources.PORTFORWARDING)

        self._connection = n_rpc.Connection()
        endpoints = [resources_rpc.ResourcesPushRpcCallback()]
        topic = resources_rpc.resource_type_versioned_topic(
            resources.PORTFORWARDING)
        self._connection.create_consumer(topic, endpoints, fanout=True)
        self._connection.consume_in_threads()

    def consume_api(self, agent_api):
        self.agent_api = agent_api

    def _handle_notification(self, context, resource_type,
                             forwardings, event_type):
        for forwarding in forwardings:
            self._process_port_forwarding_event(
                context, forwarding, event_type)

    def _store_local(self, pf_objs, event_type):
        if event_type == events.CREATED:
            self.mapping.set_port_forwardings(pf_objs)
        elif event_type == events.UPDATED:
            self.mapping.update_port_forwardings(pf_objs)
        elif event_type == events.DELETED:
            self.mapping.del_port_forwardings(pf_objs)

    def _get_fip_rules(self, port_forward, wrap_name):
        chain_rule_list = []
        pf_chain_name = self._get_port_forwarding_chain_name(port_forward.id)
        chain_rule_list.append((DEFAULT_PORT_FORWARDING_CHAIN,
                                '-j %s-%s' %
                                (wrap_name, pf_chain_name)))
        floating_ip_address = str(port_forward.floating_ip_address)
        protocol = port_forward.protocol
        internal_ip_address = str(port_forward.internal_ip_address)
        internal_port = port_forward.internal_port
        external_port = port_forward.external_port
        chain_rule = (pf_chain_name,
                      '-d %s/32 -p %s -m %s --dport %s '
                      '-j DNAT --to-destination %s:%s' % (
                          floating_ip_address, protocol, protocol,
                          external_port, internal_ip_address,
                          internal_port))
        chain_rule_list.append(chain_rule)
        return chain_rule_list

    def _rule_apply(self, iptables_manager, port_forwarding, rule_tag):
        iptables_manager.ipv4['nat'].clear_rules_by_tag(rule_tag)
        if DEFAULT_PORT_FORWARDING_CHAIN not in iptables_manager.ipv4[
             'nat'].chains:
            self._install_default_rules(iptables_manager)

        for chain, rule in self._get_fip_rules(
                port_forwarding, iptables_manager.wrap_name):
            if chain not in iptables_manager.ipv4['nat'].chains:
                iptables_manager.ipv4['nat'].add_chain(chain)
            iptables_manager.ipv4['nat'].add_rule(chain, rule, tag=rule_tag)

    @coordination.synchronized('router-lock-ns-{namespace}')
    def _process_create(self, port_forwardings, ri, interface_name, namespace,
                        iptables_manager):
        if not port_forwardings:
            return
        device = ip_lib.IPDevice(interface_name, namespace=namespace)

        is_distributed = ri.router.get('distributed')
        ha_port = ri.router.get(constants.HA_INTERFACE_KEY, None)
        fip_statuses = {}
        for port_forwarding in port_forwardings:
            # check if the port forwarding is managed in this agent from
            # OVO and router rpc.
            if port_forwarding.id in self.mapping.managed_port_forwardings:
                LOG.debug("Skip port forwarding %s for create, as it had been "
                          "managed by agent", port_forwarding.id)
                continue
            existing_cidrs = ri.get_router_cidrs(device)
            fip_ip = str(port_forwarding.floating_ip_address)
            fip_cidr = str(netaddr.IPNetwork(fip_ip))
            status = ''
            if fip_cidr not in existing_cidrs:
                try:
                    if not is_distributed:
                        fip_statuses[port_forwarding.floatingip_id] = (
                            ri.add_floating_ip(
                                {'floating_ip_address': fip_ip},
                                interface_name, device))
                    else:
                        if not ha_port:
                            device.addr.add(fip_cidr)
                            ip_lib.send_ip_addr_adv_notif(namespace,
                                                          interface_name,
                                                          fip_ip)
                        else:
                            ri._add_vip(fip_cidr, interface_name)
                        status = constants.FLOATINGIP_STATUS_ACTIVE
                except Exception:
                    # Any error will causes the fip status to be set 'ERROR'
                    status = constants.FLOATINGIP_STATUS_ERROR
                    LOG.warning("Unable to configure floating IP %(fip_id)s "
                                "for port forwarding %(pf_id)s",
                                {'fip_id': port_forwarding.floatingip_id,
                                 'pf_id': port_forwarding.id})
            else:
                if not ha_port:
                    ip_lib.send_ip_addr_adv_notif(namespace,
                                                  interface_name,
                                                  fip_ip)
            if status:
                fip_statuses[port_forwarding.floatingip_id] = status

        if ha_port and ha_port['status'] == constants.PORT_STATUS_ACTIVE:
            ri.enable_keepalived()

        for port_forwarding in port_forwardings:
            rule_tag = PORT_FORWARDING_PREFIX + port_forwarding.id
            self._rule_apply(iptables_manager, port_forwarding, rule_tag)

        iptables_manager.apply()
        self._sending_port_forwarding_fip_status(ri, fip_statuses)
        self._store_local(port_forwardings, events.CREATED)

    def _sending_port_forwarding_fip_status(self, ri, statuses):
        if not statuses:
            return
        LOG.debug('Sending Port Forwarding floating ip '
                  'statuses: %s', statuses)
        # Update floating IP status on the neutron server
        ri.agent.plugin_rpc.update_floatingip_statuses(
            ri.agent.context, ri.router_id, statuses)

    def _get_resource_by_router(self, ri):
        is_distributed = ri.router.get('distributed')
        ex_gw_port = ri.get_ex_gw_port()
        if not is_distributed:
            interface_name = ri.get_external_device_interface_name(ex_gw_port)
            namespace = ri.ns_name
            iptables_manager = ri.iptables_manager
        else:
            interface_name = ri.get_snat_external_device_interface_name(
                ex_gw_port)
            namespace = ri.snat_namespace.name
            iptables_manager = ri.snat_iptables_manager
        return interface_name, namespace, iptables_manager

    def _check_if_need_process(self, ri, force=False):
        # force means the request comes from, if True means it comes from OVO,
        # as we get a actually port forwarding object, then we need to check in
        # the following steps. But False, means it comes from router rpc.
        if not ri or not ri.get_ex_gw_port() or (
                not force and not ri.fip_managed_by_port_forwardings):
            # agent doesn't hold the router. pass
            # This router doesn't own a gw port. pass
            # This router doesn't hold a port forwarding mapping. pass
            return False

        is_distributed = ri.router.get('distributed')
        agent_mode = ri.agent_conf.agent_mode
        if (is_distributed and
                agent_mode in [constants.L3_AGENT_MODE_DVR_NO_EXTERNAL,
                               constants.L3_AGENT_MODE_DVR]):
            # just support centralized cases
            return False

        if is_distributed and not ri.snat_namespace.exists():
            return False

        return True

    def _process_port_forwarding_event(self, context, port_forwarding,
                                       event_type):
        router_id = port_forwarding.router_id
        ri = self._get_router_info(router_id)
        if not self._check_if_need_process(ri, force=True):
            return

        (interface_name, namespace,
         iptables_manager) = self._get_resource_by_router(ri)

        if event_type == events.CREATED:
            self._process_create(
                [port_forwarding], ri, interface_name, namespace,
                iptables_manager)
        elif event_type == events.UPDATED:
            self._process_update([port_forwarding], iptables_manager,
                                 interface_name, namespace)
        elif event_type == events.DELETED:
            self._process_delete(
                context, [port_forwarding], ri, interface_name, namespace,
                iptables_manager)

    @coordination.synchronized('router-lock-ns-{namespace}')
    def _process_update(self, port_forwardings, iptables_manager,
                        interface_name, namespace):
        if not port_forwardings:
            return
        device = ip_lib.IPDevice(interface_name, namespace=namespace)
        for port_forwarding in port_forwardings:
            # check if port forwarding change from OVO and router rpc
            if not self.mapping.check_port_forwarding_changes(port_forwarding):
                LOG.debug("Skip port forwarding %s for update, as there is no "
                          "difference between the memory managed by agent",
                          port_forwarding.id)
                continue
            current_chain = self._get_port_forwarding_chain_name(
                port_forwarding.id)
            iptables_manager.ipv4['nat'].remove_chain(current_chain)
            ori_pf = self.mapping.managed_port_forwardings[port_forwarding.id]
            device.delete_socket_conntrack_state(
                str(ori_pf.floating_ip_address), ori_pf.external_port,
                protocol=ori_pf.protocol)
            rule_tag = PORT_FORWARDING_PREFIX + port_forwarding.id
            self._rule_apply(iptables_manager, port_forwarding, rule_tag)
        iptables_manager.apply()
        self._store_local(port_forwardings, events.UPDATED)

    @coordination.synchronized('router-lock-ns-{namespace}')
    def _process_delete(self, context, port_forwardings, ri, interface_name,
                        namespace, iptables_manager):
        if not port_forwardings:
            return
        device = ip_lib.IPDevice(interface_name, namespace=namespace)
        for port_forwarding in port_forwardings:
            current_chain = self._get_port_forwarding_chain_name(
                port_forwarding.id)
            iptables_manager.ipv4['nat'].remove_chain(current_chain)
            fip_address = str(port_forwarding.floating_ip_address)
            device.delete_socket_conntrack_state(
                fip_address, port_forwarding.external_port,
                protocol=port_forwarding.protocol)

        iptables_manager.apply()

        fip_id_cidrs = set([(pf.floatingip_id,
                             str(netaddr.IPNetwork(pf.floating_ip_address)))
                            for pf in port_forwardings])
        self._sync_and_remove_fip(context, fip_id_cidrs, device, ri)
        self._store_local(port_forwardings, events.DELETED)

    def _sync_and_remove_fip(self, context, fip_id_cidrs, device, ri):
        if not fip_id_cidrs:
            return
        ha_port = ri.router.get(constants.HA_INTERFACE_KEY)
        fip_ids = [item[0] for item in fip_id_cidrs]
        pfs = self.resource_rpc.bulk_pull(context, resources.PORTFORWARDING,
                                          filter_kwargs={
                                              'floatingip_id': fip_ids})
        exist_fips = set()
        fip_status = {}
        for pf in pfs:
            exist_fips.add(pf.floatingip_id)

        for fip_id_cidr in fip_id_cidrs:
            if fip_id_cidr[0] not in exist_fips:
                if ha_port:
                    ri._remove_vip(fip_id_cidr[1])
                else:
                    device.delete_addr_and_conntrack_state(fip_id_cidr[1])
                fip_status[fip_id_cidr[0]] = 'DOWN'

        if ha_port:
            ri.enable_keepalived()
        self._sending_port_forwarding_fip_status(ri, fip_status)
        for fip_id in fip_status.keys():
            self.mapping.clear_by_fip(fip_id, ri.router_id)

    def _get_router_info(self, router_id):
        router_info = self.agent_api.get_router_info(router_id)
        if router_info:
            return router_info
        LOG.debug("Router %s is not managed by this agent. "
                  "It was possibly deleted concurrently.", router_id)

    def _get_port_forwarding_chain_name(self, pf_id):
        chain_name = PORT_FORWARDING_CHAIN_PREFIX + pf_id
        return chain_name[:constants.MAX_IPTABLES_CHAIN_LEN_WRAP]

    def _install_default_rules(self, iptables_manager):
        default_rule = '-j %s-%s' % (iptables_manager.wrap_name,
                                     DEFAULT_PORT_FORWARDING_CHAIN)
        iptables_manager.ipv4['nat'].add_chain(DEFAULT_PORT_FORWARDING_CHAIN)
        iptables_manager.ipv4['nat'].add_rule('PREROUTING', default_rule)
        iptables_manager.apply()

    def check_local_port_forwardings(self, context, ri, fip_ids):
        pfs = self.resource_rpc.bulk_pull(context, resources.PORTFORWARDING,
                                          filter_kwargs={
                                              'floatingip_id': fip_ids})

        (interface_name, namespace,
         iptable_manager) = self._get_resource_by_router(ri)
        local_pfs = set(self.mapping.managed_port_forwardings.keys())
        new_pfs = []
        updated_pfs = []
        current_pfs = set()
        for pf in pfs:
            # check the request port forwardings, and split them into
            # update, new, current part from router rpc
            if pf.id in self.mapping.managed_port_forwardings:
                if self.mapping.check_port_forwarding_changes(pf):
                    updated_pfs.append(pf)
            else:
                new_pfs.append(pf)
            current_pfs.add(pf.id)
        remove_pf_ids_set = local_pfs - current_pfs
        remove_pfs = [self.mapping.managed_port_forwardings[pf_id]
                      for pf_id in remove_pf_ids_set]
        self._process_update(updated_pfs, iptable_manager,
                             interface_name, namespace)
        self._process_create(new_pfs, ri, interface_name,
                             namespace, iptable_manager)
        self._process_delete(context, remove_pfs, ri, interface_name,
                             namespace, iptable_manager)

    def process_port_forwarding(self, context, data):
        ri = self._get_router_info(data['id'])
        if not self._check_if_need_process(ri):
            return

        self.check_local_port_forwardings(
            context, ri, ri.fip_managed_by_port_forwardings)

    def add_router(self, context, data):
        """Handle a router add event.

        Called on router create.

        :param context: RPC context.
        :param data: Router data.
        """
        self.process_port_forwarding(context, data)

    def update_router(self, context, data):
        """Handle a router update event.

        Called on router update.

        :param context: RPC context.
        :param data: Router data.
        """
        self.process_port_forwarding(context, data)

    def delete_router(self, context, data):
        """Handle a router delete event.

        :param context: RPC context.
        :param data: Router data.
        """
        self.mapping.clean_port_forwardings_by_router_id(data['id'])

    def ha_state_change(self, context, data):
        pass

    def update_network(self, context, data):
        pass
